# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import numpy as np
import pytest
import torch
from scipy.optimize import linear_sum_assignment as scipy_linear_sum_assignment

from nemo.collections.asr.data.audio_to_label import repeat_signal
from nemo.collections.asr.parts.utils.longform_clustering import LongFormSpeakerClustering
from nemo.collections.asr.parts.utils.offline_clustering import (
    SpeakerClustering,
    get_scale_interpolated_embs,
    getCosAffinityMatrix,
    getKneighborsConnections,
    split_input_data,
)
from nemo.collections.asr.parts.utils.online_clustering import (
    OnlineSpeakerClustering,
    get_closest_embeddings,
    get_merge_quantity,
    get_minimal_indices,
    merge_vectors,
    run_reducer,
    stitch_cluster_labels,
)
from nemo.collections.asr.parts.utils.optimization_utils import LinearSumAssignmentSolver
from nemo.collections.asr.parts.utils.optimization_utils import linear_sum_assignment as nemo_linear_sum_assignment
from nemo.collections.asr.parts.utils.speaker_utils import (
    OnlineSegmentor,
    check_ranges,
    fl2int,
    get_new_cursor_for_update,
    get_online_segments_from_slices,
    get_online_subsegments_from_buffer,
    get_speech_labels_for_update,
    get_sub_range_list,
    get_subsegments,
    get_subsegments_scriptable,
    get_target_sig,
    int2fl,
    is_overlap,
    merge_float_intervals,
    merge_int_intervals,
    tensor_to_list,
)


def check_range_values(target, source):
    bool_list = []
    for tgt, src in zip(target, source):
        for x, y in zip(src, tgt):
            bool_list.append(abs(x - y) < 1e-6)
    return all(bool_list)


def check_labels(target, source):
    bool_list = []
    for x, y in zip(target, source):
        bool_list.append(abs(x - y) < 1e-6)
    return all(bool_list)


def matrix(mat, use_tensor=True, dtype=torch.long):
    if use_tensor:
        mat = torch.Tensor(mat).to(dtype)
    else:
        mat = np.array(mat)
    return mat


def generate_orthogonal_embs(total_spks, perturb_sigma, emb_dim):
    """Generate a set of artificial orthogonal embedding vectors from random numbers"""
    gaus = torch.randn(emb_dim, emb_dim)
    _svd = torch.linalg.svd(gaus)
    orth = _svd[0] @ _svd[2]
    orth_embs = orth[:total_spks]
    # Assert orthogonality
    assert torch.abs(getCosAffinityMatrix(orth_embs) - torch.diag(torch.ones(total_spks))).sum() < 1e-4
    return orth_embs


def generate_toy_data(
    n_spks=2,
    spk_dur=3,
    emb_dim=192,
    perturb_sigma=0.0,
    ms_window=[1.5, 1.0, 0.5],
    ms_shift=[0.75, 0.5, 0.25],
    torch_seed=0,
):
    torch.manual_seed(torch_seed)
    spk_timestamps = [(spk_dur * k, spk_dur) for k in range(n_spks)]
    emb_list, seg_list = [], []
    multiscale_segment_counts = [0 for _ in range(len(ms_window))]
    ground_truth = []
    random_orthogonal_embs = generate_orthogonal_embs(n_spks, perturb_sigma, emb_dim)
    for scale_idx, (window, shift) in enumerate(zip(ms_window, ms_shift)):
        for spk_idx, (offset, dur) in enumerate(spk_timestamps):
            segments_stt_dur = get_subsegments_scriptable(offset=offset, window=window, shift=shift, duration=dur)
            segments = [[x[0], x[0] + x[1]] for x in segments_stt_dur]
            emb_cent = random_orthogonal_embs[spk_idx, :]
            emb = emb_cent.tile((len(segments), 1)) + 0.1 * torch.rand(len(segments), emb_dim)
            seg_list.extend(segments)
            emb_list.append(emb)
            if emb.shape[0] == 0:
                import ipdb

                ipdb.set_trace()
            multiscale_segment_counts[scale_idx] += emb.shape[0]

            if scale_idx == len(multiscale_segment_counts) - 1:
                ground_truth.extend([spk_idx] * emb.shape[0])

    emb_tensor = torch.concat(emb_list)
    multiscale_segment_counts = torch.tensor(multiscale_segment_counts)
    segm_tensor = torch.tensor(seg_list)
    multiscale_weights = torch.ones(len(ms_window)).unsqueeze(0)
    ground_truth = torch.tensor(ground_truth)
    return emb_tensor, segm_tensor, multiscale_segment_counts, multiscale_weights, spk_timestamps, ground_truth


class TestDiarizationSequneceUtilFunctions:
    """Tests diarization and speaker-task related utils."""

    @pytest.mark.unit
    @pytest.mark.parametrize("Y", [[3, 3, 3, 4, 4, 5], [100, 100, 100, 104, 104, 1005]])
    @pytest.mark.parametrize("target", [[0, 0, 0, 1, 1, 2]])
    @pytest.mark.parametrize("offset", [1, 10])
    def test_minimal_index_ex2(self, Y, target, offset):
        Y = torch.tensor(Y)
        target = torch.tensor(target)
        min_Y = get_minimal_indices(Y)
        assert check_labels(target, min_Y)
        min_Y = get_minimal_indices(Y + offset)
        assert check_labels(target, min_Y)

    @pytest.mark.parametrize("Y", [[4, 0, 0, 5, 4, 5], [14, 12, 12, 19, 14, 19]])
    @pytest.mark.parametrize("target", [[1, 0, 0, 2, 1, 2]])
    @pytest.mark.parametrize("offset", [1, 10])
    def test_minimal_index_ex2(self, Y, target, offset):
        Y = torch.tensor(Y)
        target = torch.tensor(target)
        min_Y = get_minimal_indices(Y)
        assert check_labels(target, min_Y)
        min_Y = get_minimal_indices(Y + offset)
        assert check_labels(target, min_Y)

    @pytest.mark.unit
    @pytest.mark.parametrize("N", [2, 4, 16, 64])
    def test_minimal_index_same(self, N):
        Y = matrix([0] * N + [1] * N + [2] * N)
        min_Y = get_minimal_indices(Y)
        target = matrix([0] * N + [1] * N + [2] * N)
        assert check_labels(target, min_Y)

    @pytest.mark.unit
    @pytest.mark.parametrize("N", [2, 4, 16, 64])
    def test_stitch_cluster_labels_label_switch(self, N):
        Y_old = matrix([0] * N)
        Y_new = matrix([0] * N) + 1
        target = matrix([0] * N)
        result = stitch_cluster_labels(Y_old, Y_new)
        assert check_labels(target, result)

    @pytest.mark.unit
    @pytest.mark.parametrize("N", [2, 4, 16, 64])
    def test_stitch_cluster_labels_label_many_to_one(self, N):
        Y_old = matrix(np.arange(N).tolist())
        Y_new = matrix([0] * N)
        target = matrix([0] * N)
        result = stitch_cluster_labels(Y_old, Y_new)
        assert check_labels(target, result)

    @pytest.mark.unit
    @pytest.mark.parametrize("N", [2, 4, 16, 64])
    def test_stitch_cluster_labels_label_one_to_many(self, N):
        Y_old = matrix(np.arange(N).tolist())
        Y_new = matrix([k for k in range(N)])
        target = matrix([k for k in range(N)])
        result = stitch_cluster_labels(Y_old, Y_new)
        assert check_labels(target, result)

    @pytest.mark.unit
    @pytest.mark.parametrize("N", [2, 4, 16, 64])
    def test_stitch_cluster_labels_one_label_replaced(self, N):
        Y_old = matrix([0] * N + [1] * N + [2] * N)
        Y_new = matrix([1] * N + [2] * N + [3] * N)
        target = matrix([0] * N + [1] * N + [2] * N)
        result = stitch_cluster_labels(Y_old, Y_new)
        assert check_labels(target, result)

    @pytest.mark.unit
    @pytest.mark.parametrize("N", [2, 4, 16, 64])
    def test_stitch_cluster_labels_confusion_error(self, N):
        Y_old = matrix([0] * N + [1] * (N - 1) + [2] * (N + 1))
        Y_new = matrix([1] * N + [2] * N + [3] * N)
        target = matrix([0] * N + [1] * N + [2] * N)
        result = stitch_cluster_labels(Y_old, Y_new)
        assert check_labels(target, result)

    @pytest.mark.unit
    @pytest.mark.parametrize("N", [2, 256])
    def test_stitch_cluster_labels_speaker_more_speakers(self, N):
        Y_old = matrix([0] * N + [1] * (N - 1) + [2] * (N + 1) + [0, 0, 0])
        Y_new = matrix([1] * N + [0] * N + [2] * N + [4, 5, 6])
        target = matrix([0] * N + [1] * N + [2] * N + [3, 4, 5])
        result = stitch_cluster_labels(Y_old, Y_new)
        assert check_labels(target, result)

    @pytest.mark.unit
    @pytest.mark.parametrize("N", [2, 256])
    def test_stitch_cluster_labels_speaker_longer_sequence(self, N):
        Y_old = matrix([0] * N + [1] * N + [2] * N + [0, 0, 0] * N)
        Y_new = matrix([1] * N + [2] * N + [0] * N + [1, 2, 3, 1, 2, 3] * N)
        target = matrix([0] * N + [1] * N + [2] * N + [0, 1, 3, 0, 1, 3] * N)
        result = stitch_cluster_labels(Y_old, Y_new)
        assert check_labels(target, result)

    @pytest.mark.unit
    @pytest.mark.parametrize("n_spks", [2, 3, 4, 5])
    @pytest.mark.parametrize("merge_quantity", [2, 3])
    def test_embedding_merger(self, n_spks, merge_quantity):
        em, ts, mc, mw, spk_ts, gt = generate_toy_data(n_spks, spk_dur=5, perturb_sigma=10)
        em_s, ts_s = split_input_data(em, ts, mc)
        target_speaker_index = 0
        pre_clus_labels = gt
        ndx = torch.where(pre_clus_labels == target_speaker_index)[0]
        pre_embs = em_s[-1]
        affinity_mat = getCosAffinityMatrix(pre_embs)
        cmat = affinity_mat[:, ndx][ndx, :]
        # Check the dimension of the selected affinity values
        assert cmat.shape[0] == cmat.shape[1] == torch.sum(pre_clus_labels == target_speaker_index).item()
        index_2d, rest_inds = get_closest_embeddings(cmat, merge_quantity)
        # Check the most closest affinity value
        assert torch.max(cmat.sum(0)) == cmat.sum(0)[index_2d[0]]
        spk_cluster_labels, emb_ndx = pre_clus_labels[ndx], pre_embs[ndx]
        merged_embs, merged_clus_labels = merge_vectors(index_2d, emb_ndx, spk_cluster_labels)
        # Check the number of merged embeddings and labels
        assert (torch.sum(gt == target_speaker_index).item() - merge_quantity) == merged_clus_labels.shape[0]

    @pytest.mark.unit
    @pytest.mark.parametrize("n_spks", [1, 8])
    @pytest.mark.parametrize("spk_dur", [0.2, 0.25, 0.5, 1, 10])
    def test_cosine_affinity_calculation(self, n_spks, spk_dur):
        em, ts, mc, mw, spk_ts, gt = generate_toy_data(n_spks=n_spks, spk_dur=spk_dur)
        em_s, ts_s = split_input_data(em, ts, mc)
        affinity_mat = getCosAffinityMatrix(em_s[-1])
        # affinity_mat should not contain any nan element
        assert torch.any(torch.isnan(affinity_mat)) == False

    @pytest.mark.unit
    @pytest.mark.parametrize("n_spks", [1, 8])
    @pytest.mark.parametrize("spk_dur", [0.2, 0.25, 0.5, 1, 10])
    def test_cosine_affinity_calculation_scale_interpol(self, n_spks, spk_dur):
        em, ts, mc, mw, spk_ts, gt = generate_toy_data(n_spks=n_spks, spk_dur=spk_dur)
        em_s, ts_s = split_input_data(em, ts, mc)
        embs, _ = get_scale_interpolated_embs(mw, em_s, ts_s)
        affinity_mat = getCosAffinityMatrix(embs)
        # affinity_mat should not contain any nan element
        assert torch.any(torch.isnan(affinity_mat)) == False

    @pytest.mark.unit
    @pytest.mark.parametrize("n_spks", [4, 5, 6])
    @pytest.mark.parametrize("target_speaker_index", [0, 1, 2])
    @pytest.mark.parametrize("merge_quantity", [2, 3])
    def test_embedding_reducer(self, n_spks, target_speaker_index, merge_quantity):
        em, ts, mc, mw, spk_ts, gt = generate_toy_data(n_spks=n_spks, spk_dur=10)
        em_s, ts_s = split_input_data(em, ts, mc)
        merged_embs, merged_clus_labels, _ = run_reducer(
            pre_embs=em_s[-1],
            target_spk_idx=target_speaker_index,
            merge_quantity=merge_quantity,
            pre_clus_labels=gt,
        )
        assert (torch.sum(gt == target_speaker_index).item() - merge_quantity) == merged_clus_labels.shape[0]

    @pytest.mark.unit
    @pytest.mark.parametrize("ntbr", [3])
    @pytest.mark.parametrize("pcl", [torch.tensor([0] * 70 + [1] * 32)])
    @pytest.mark.parametrize("mspb", [25])
    def test_merge_scheduler_2clus(self, ntbr, pcl, mspb):
        class_target_vol = get_merge_quantity(
            num_to_be_removed=ntbr,
            pre_clus_labels=pcl,
            min_count_per_cluster=mspb,
        )
        assert all(class_target_vol == torch.tensor([3, 0]))

    @pytest.mark.unit
    @pytest.mark.parametrize("ntbr", [3])
    @pytest.mark.parametrize("pcl", [torch.tensor([0] * 80 + [1] * 35 + [2] * 32)])
    @pytest.mark.parametrize("mspb", [0, 25])
    def test_merge_scheduler_3clus(self, ntbr, pcl, mspb):
        class_target_vol = get_merge_quantity(
            num_to_be_removed=ntbr,
            pre_clus_labels=pcl,
            min_count_per_cluster=mspb,
        )
        assert all(class_target_vol == torch.tensor([3, 0, 0]))

    @pytest.mark.unit
    @pytest.mark.parametrize("ntbr", [132 - 45])
    @pytest.mark.parametrize("pcl", [torch.tensor([2] * 70 + [0] * 32 + [1] * 27 + [3] * 3)])
    @pytest.mark.parametrize("mspb", [3, 10])
    def test_merge_scheduler_4clus_shuff(self, ntbr, pcl, mspb):
        class_target_vol = get_merge_quantity(
            num_to_be_removed=ntbr,
            pre_clus_labels=pcl,
            min_count_per_cluster=mspb,
        )
        assert all(class_target_vol == torch.tensor([18, 13, 56, 0]))

    @pytest.mark.unit
    @pytest.mark.parametrize("ntbr", [3])
    @pytest.mark.parametrize("pcl", [torch.tensor([0] * 5 + [1] * 4 + [2] * 3)])
    @pytest.mark.parametrize("mspb", [0, 2])
    def test_merge_scheduler_3clus(self, ntbr, pcl, mspb):
        class_target_vol = get_merge_quantity(
            num_to_be_removed=ntbr,
            pre_clus_labels=pcl,
            min_count_per_cluster=mspb,
        )
        assert all(class_target_vol == torch.tensor([2, 1, 0]))

    @pytest.mark.unit
    @pytest.mark.parametrize("ntbr", [2])
    @pytest.mark.parametrize("pcl", [torch.tensor([0] * 7 + [1] * 5 + [2] * 3 + [3] * 5)])
    @pytest.mark.parametrize("mspb", [2])
    def test_merge_scheduler_3clus_repeat(self, ntbr, pcl, mspb):
        class_target_vol = get_merge_quantity(
            num_to_be_removed=ntbr,
            pre_clus_labels=pcl,
            min_count_per_cluster=mspb,
        )
        assert all(class_target_vol == torch.tensor([2, 0, 0, 0]))


class TestClassExport:
    @pytest.mark.unit
    def test_online_segmentor_class_export(self):
        _OnlineSegmentor = torch.jit.script(OnlineSegmentor)
        online_segmentor = _OnlineSegmentor(sample_rate=16000)
        assert isinstance(online_segmentor, OnlineSegmentor)

    @pytest.mark.unit
    def test_online_segmentor_instance_export(self):
        online_segmentor = OnlineSegmentor(sample_rate=16000)
        online_segmentor = torch.jit.script(online_segmentor)
        isinstance(online_segmentor, torch.jit._script.RecursiveScriptClass)

    @pytest.mark.unit
    def test_online_speaker_clustering_instance_export(self):
        online_clus = OnlineSpeakerClustering(
            max_num_speakers=8,
            max_rp_threshold=0.15,
            sparse_search_volume=30,
            history_buffer_size=150,
            current_buffer_size=150,
            cuda=True,
        )
        online_clus = torch.jit.script(online_clus)
        isinstance(online_clus, torch.jit._script.RecursiveScriptClass)

    @pytest.mark.unit
    def test_online_speaker_clustering_instance_export(self):
        offline_speaker_clustering = SpeakerClustering(maj_vote_spk_count=False, min_samples_for_nmesc=0, cuda=True)
        offline_speaker_clustering = torch.jit.script(offline_speaker_clustering)
        isinstance(offline_speaker_clustering, torch.jit._script.RecursiveScriptClass)


class TestGetSubsegments:
    @pytest.mark.unit
    @pytest.mark.parametrize(
        "offset, window, shift, duration, min_subsegment_duration, decimals, use_asr_style_frame_count, sample_rate, feat_per_sec, expected",
        [
            (12.05, 1.5, 0.75, 2.4, 0.01, 2, False, 16000, 100, [[12.05, 1.5], [12.8, 1.5], [13.55, 0.9]]),
            (0, 1.0, 0.5, 0.4, 0.01, 2, False, 16000, 100, [[0, 0.4]]),
            (0, 2.0, 1.0, 1.5, 0.5, 2, False, 16000, 100, [[0, 1.5]]),
            (
                10,
                1.5,
                0.75,
                4.5,
                0.5,
                2,
                False,
                16000,
                100,
                [[10, 1.5], [10.75, 1.5], [11.5, 1.5], [12.25, 1.5], [13.0, 1.5]],
            ),
            (0, 1.5, 0.5, 0.3, 0.01, 2, True, 16000, 100, [[0, 0.3]]),
        ],
    )
    def test_get_subsegments(
        self,
        offset,
        window,
        shift,
        duration,
        min_subsegment_duration,
        decimals,
        use_asr_style_frame_count,
        sample_rate,
        feat_per_sec,
        expected,
    ):

        for is_scriptable in [True, False]:
            if is_scriptable:
                result = get_subsegments_scriptable(
                    offset=offset,
                    window=window,
                    shift=shift,
                    duration=duration,
                )
            else:
                result = get_subsegments(
                    offset=offset,
                    window=window,
                    shift=shift,
                    duration=duration,
                    min_subsegment_duration=min_subsegment_duration,
                    decimals=decimals,
                    use_asr_style_frame_count=use_asr_style_frame_count,
                    sample_rate=sample_rate,
                    feat_per_sec=feat_per_sec,
                )
            result_round = []
            for subsegment in result:
                result_round.append([round(x, decimals) for x in subsegment])
            assert result_round == expected

    @pytest.mark.unit
    def test_min_subsegment_duration_filtering(self):
        result = get_subsegments(
            offset=0,
            window=1.5,
            shift=0.5,
            duration=3,
            min_subsegment_duration=2.0,
            decimals=2,
            use_asr_style_frame_count=False,
        )
        expected = []  # Only subsegments meeting the duration filter should remain
        assert result == expected

    @pytest.mark.unit
    def test_zero_duration(self):
        result = get_subsegments(
            offset=0,
            window=1.0,
            shift=0.5,
            duration=0,
            min_subsegment_duration=0.01,
            decimals=2,
            use_asr_style_frame_count=False,
        )
        assert result == []

    @pytest.mark.unit
    def test_edge_case_short_slice(self):
        result = get_subsegments(
            offset=0,
            window=0.5,
            shift=0.25,  # Shift larger than duration
            duration=0.25,
            min_subsegment_duration=0.01,
            decimals=2,
            use_asr_style_frame_count=False,
        )
        assert result == [[0.0, 0.25]]


class TestDiarizationSegmentationUtils:
    """
    Test segmentation util functions
    """

    @pytest.mark.unit
    @pytest.mark.parametrize(
        "intervals",
        [
            [[1, 4], [2, 6], [8, 10], [15, 18]],
            [[8, 10], [15, 18], [2, 6], [1, 3]],
            [[8, 10], [15, 18], [2, 6], [1, 3], [3, 5]],
            [[8, 10], [8, 8], [15, 18], [2, 6], [1, 6], [2, 4]],
        ],
    )
    @pytest.mark.parametrize("target", [[[1, 6], [8, 10], [15, 18]]])
    def test_merge_int_intervals_ex1(self, intervals, target):
        merged = merge_int_intervals(intervals)
        assert check_range_values(target, merged)

    @pytest.mark.unit
    @pytest.mark.parametrize(
        "intervals",
        [
            [[6, 8], [0, 9], [2, 4], [4, 7]],
            [[0, 9], [6, 8], [4, 7], [2, 4]],
            [[0, 4], [0, 0], [4, 9], [2, 4]],
            [[6, 8], [2, 8], [0, 3], [3, 4], [4, 5], [5, 9]],
        ],
    )
    @pytest.mark.parametrize("target", [[[0, 9]]])
    def test_merge_int_intervals_ex2(self, intervals, target):
        merged = merge_int_intervals(intervals)
        assert check_range_values(target, merged)

    @pytest.mark.unit
    @pytest.mark.parametrize("intervals", [[[0, 1], [1, 9]], [[0, 0], [0, 9]], [[0, 9], [0, 9]]])
    @pytest.mark.parametrize("target", [[[0, 9]]])
    def test_merge_int_intervals_edge_test(self, intervals, target):
        merged = merge_int_intervals(intervals)
        assert check_range_values(target, merged)

    @pytest.mark.unit
    @pytest.mark.parametrize("rangeA", [[1.0, 2.0]])
    @pytest.mark.parametrize("rangeB", [[0.5, 1.5], [0.9999, 1.0001]])
    def test_is_overlap_true(self, rangeA, rangeB):
        assert is_overlap(rangeA, rangeB)

    @pytest.mark.unit
    @pytest.mark.parametrize("rangeA", [[1.0, 2.0]])
    @pytest.mark.parametrize("rangeB", [[2.0, 2.5], [-1.0, 1.00]])
    def test_is_overlap_false(self, rangeA, rangeB):
        assert not is_overlap(rangeA, rangeB)

    @pytest.mark.unit
    @pytest.mark.parametrize("x", [1.0, 2.3456])
    @pytest.mark.parametrize("decimals", [1, 2, 3, 4])
    def test_fl2int(self, x, decimals):
        assert fl2int(x, decimals) == round(x * 10**decimals, 0)

    @pytest.mark.unit
    @pytest.mark.parametrize("x", [1234])
    @pytest.mark.parametrize(
        "decimals",
        [
            1,
            2,
            3,
            4,
        ],
    )
    def test_int2fl(self, x, decimals):
        assert abs(int2fl(x, decimals) - round(x / (10**decimals), decimals)) < (10 ** -(decimals + 1))

    @pytest.mark.unit
    def test_merge_float_intervals_edge_margin_test(self):
        intervals = [[0.0, 1.0], [1.0, 2.0]]

        target_0 = [[0.0, 2.0]]
        merged_0 = merge_float_intervals(intervals, margin=0)
        assert check_range_values(target_0, merged_0)

        target_1 = [[0.0, 1.0], [1.0, 2.0]]
        merged_1 = merge_float_intervals(intervals, margin=1)
        assert check_range_values(target_1, merged_1)

        target_2 = [[0.0, 1.0], [1.0, 2.0]]
        merged_2 = merge_float_intervals(intervals, margin=2)
        assert check_range_values(target_2, merged_2)

    @pytest.mark.unit
    @pytest.mark.parametrize(
        "intervals",
        [
            [[0.25, 1.7], [1.5, 3.0], [2.8, 5.0], [5.5, 10.0]],
            [[0.25, 5.0], [5.5, 10.0], [1.5, 3.5]],
            [[5.5, 8.05], [8.0, 10.0], [0.25, 5.0]],
            [[0.25, 3.0], [1.5, 3.0], [5.5, 10.0], [2.8, 5.0]],
            [[0.25, 1.7], [1.5, 3.0], [2.8, 5.0], [5.5, 10.0]],
        ],
    )
    @pytest.mark.parametrize("target", [[[0.25, 5.0], [5.5, 10.0]]])
    def test_merge_float_overlaps(self, intervals, target):
        merged = merge_float_intervals(intervals)
        assert check_range_values(target, merged)

    @pytest.mark.unit
    def test_get_speech_labels_for_update(self):
        frame_start = 3.0
        buffer_end = 6.0
        cumulative_speech_labels = torch.tensor([[0.0000, 3.7600]])
        vad_timestamps = torch.tensor([[0.9600, 4.8400]])
        cursor_for_old_segments = 1.0
        speech_labels_for_update, cumulative_speech_labels = get_speech_labels_for_update(
            frame_start,
            buffer_end,
            cumulative_speech_labels,
            vad_timestamps,
            cursor_for_old_segments,
        )
        assert (speech_labels_for_update - torch.tensor([[1.0000, 3.7600]])).sum() < 1e-8
        assert (cumulative_speech_labels - torch.tensor([[0.9600, 4.8400]])).sum() < 1e-8

        # Check if the ranges are containing faulty values
        assert check_ranges(speech_labels_for_update)
        assert check_ranges(cumulative_speech_labels)

    @pytest.mark.unit
    def test_get_online_subsegments_from_buffer(self):
        torch.manual_seed(0)
        sample_rate = 16000
        speech_labels_for_update = torch.Tensor([[0.0000, 3.7600]])
        audio_buffer = torch.randn(5 * sample_rate)
        segment_indexes = []
        window = 2.0
        shift = 1.0
        slice_length = int(window * sample_rate)
        range_target = [[0.0, 2.0], [1.0, 3.0], [2.0, 3.76]]
        sigs_list, sig_rangel_list, sig_indexes = get_online_subsegments_from_buffer(
            buffer_start=0.0,
            buffer_end=5.0,
            sample_rate=sample_rate,
            speech_labels_for_update=speech_labels_for_update,
            audio_buffer=audio_buffer,
            segment_indexes=segment_indexes,
            window=window,
            shift=shift,
        )
        assert check_range_values(target=range_target, source=sig_rangel_list)
        for k, rg in enumerate(sig_rangel_list):
            signal = get_target_sig(audio_buffer, rg[0], rg[1], slice_length, sample_rate)
            if len(signal) < int(window * sample_rate):
                signal = repeat_signal(signal, len(signal), slice_length)
            assert len(signal) == int(slice_length), "Length mismatch"
            assert (np.abs(signal - sigs_list[k])).sum() < 1e-8, "Audio stream mismatch"
        assert (torch.tensor(sig_indexes) - torch.arange(len(range_target))).sum() < 1e-8, "Segment index mismatch"

    @pytest.mark.unit
    @pytest.mark.parametrize("frame_start", [3.0])
    @pytest.mark.parametrize("segment_range_ts", [[[0.0, 2.0]]])
    @pytest.mark.parametrize("gt_cursor_for_old_segments", [3.0])
    @pytest.mark.parametrize("gt_cursor_index", [1])
    def test_get_new_cursor_for_update_mulsegs_ex1(
        self, frame_start, segment_range_ts, gt_cursor_for_old_segments, gt_cursor_index
    ):
        cursor_for_old_segments, cursor_index = get_new_cursor_for_update(frame_start, segment_range_ts)
        assert cursor_for_old_segments == gt_cursor_for_old_segments
        assert cursor_index == gt_cursor_index

    @pytest.mark.unit
    @pytest.mark.parametrize("target_range", [[1.0, 4.0]])
    @pytest.mark.parametrize(
        "source_range_list", [[[2.0, 3.0], [3.0, 4.0]], [[0.0, 2.0], [3.0, 5.0]], [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]]
    )
    def get_sub_range_list(self, target_range, source_range_list):
        sub_range_list = get_sub_range_list(target_range, source_range_list)
        assert sub_range_list == [[2.0, 3.0], [3.0, 4.0]]

    @pytest.mark.unit
    @pytest.mark.parametrize("source_range_list", [[[0.0, 2.0]], [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]])
    def test_tensor_to_list(self, source_range_list):
        a_range_tensor = torch.tensor(source_range_list)
        converted_list = tensor_to_list(a_range_tensor)
        assert source_range_list == converted_list

    @pytest.mark.unit
    @pytest.mark.parametrize(
        "buffer_start, buffer_end, subsegments, ind_offset, window, sample_rate",
        [
            (0.0, 2.0, [[0.5, 1.0], [1.5, 2.0]], 0, 0.1, 16000),
            (0.0, 5.0, [[0.5, 2.5], [2.7, 5.0]], 0, 1.0, 16000),
        ],
    )
    def test_get_online_segments_from_slices(
        self, buffer_start, buffer_end, subsegments, ind_offset, window, sample_rate
    ):
        sig = torch.randn(int(sample_rate * buffer_end))
        ind_offset, sigs_list, sig_rangel_list, sig_indexes = get_online_segments_from_slices(
            sig, buffer_start, buffer_end, subsegments, ind_offset, window, sample_rate
        )
        assert ind_offset == 2
        assert len(sigs_list) == 2
        assert len(sig_rangel_list) == 2
        assert len(sig_indexes) == 2


class TestClusteringUtilFunctions:
    @pytest.mark.parametrize("p_value", [1, 5, 9])
    @pytest.mark.parametrize("N", [9, 20])
    @pytest.mark.parametrize("mask_method", ['binary', 'sigmoid', 'drop'])
    def test_get_k_neighbors_connections(self, p_value: int, N: int, mask_method: str, seed=0):
        torch.manual_seed(seed)
        random_mat = torch.rand(N, N)
        affinity_mat = 0.5 * (random_mat + random_mat.T)
        affinity_mat = affinity_mat / torch.max(affinity_mat)
        binarized_affinity_mat = getKneighborsConnections(affinity_mat, p_value, mask_method)
        if mask_method == 'binary':
            assert all(binarized_affinity_mat.sum(dim=0) == float(p_value))
        elif mask_method == 'sigmoid':
            assert all(binarized_affinity_mat.sum(dim=0) <= float(p_value))
        elif mask_method == 'drop':
            assert all(binarized_affinity_mat.sum(dim=0) <= float(p_value))

    @pytest.mark.unit
    @pytest.mark.parametrize("Y_aggr", [torch.tensor([0, 1, 0, 1])])
    @pytest.mark.parametrize("chunk_cluster_count, embeddings_per_chunk", [(2, 50)])
    @pytest.mark.parametrize("window_range_list", [[[0, 1], [1, 2], [2, 3], [3, 4]]])
    @pytest.mark.parametrize(
        "absolute_merge_mapping",
        [[[torch.tensor([]), torch.tensor([0, 2])], [torch.tensor([]), torch.tensor([1, 3])]]],
    )
    @pytest.mark.parametrize("org_len", [4])
    def test_unpack_labels(
        self, Y_aggr, window_range_list, absolute_merge_mapping, chunk_cluster_count, embeddings_per_chunk, org_len
    ):
        expected_result = Y_aggr
        longform_speaker_clustering = LongFormSpeakerClustering(cuda=False)
        output = longform_speaker_clustering.unpack_labels(Y_aggr, window_range_list, absolute_merge_mapping, org_len)
        assert torch.equal(output, expected_result)


class TestSpeakerClustering:
    """
    Test speaker clustering module
    """

    @pytest.mark.unit
    @pytest.mark.parametrize("cuda", [True, False])
    def test_offline_clus_script_save_load(self, cuda):
        exported_filename = 'speaker_clustering_script.pt'
        speaker_clustering_python = SpeakerClustering(maj_vote_spk_count=False, cuda=cuda)
        speaker_clustering_scripted_source = torch.jit.script(speaker_clustering_python)
        torch.jit.save(speaker_clustering_scripted_source, exported_filename)
        assert os.path.exists(exported_filename)
        os.remove(exported_filename)
        assert not os.path.exists(exported_filename)

    @pytest.mark.unit
    @pytest.mark.parametrize("cuda", [True, False])
    def test_online_clus_script_save_load(self, cuda):
        exported_filename = 'speaker_clustering_script.pt'
        speaker_clustering_python = OnlineSpeakerClustering(
            max_num_speakers=8,
            max_rp_threshold=0.15,
            sparse_search_volume=30,
            history_buffer_size=150,
            current_buffer_size=150,
            cuda=cuda,
        )
        speaker_clustering_scripted_source = torch.jit.script(speaker_clustering_python)
        torch.jit.save(speaker_clustering_scripted_source, exported_filename)
        assert os.path.exists(exported_filename)
        os.remove(exported_filename)
        assert not os.path.exists(exported_filename)

    @pytest.mark.run_only_on('GPU')
    @pytest.mark.unit
    @pytest.mark.parametrize("n_spks", [1, 2, 3, 4, 5, 6, 7])
    @pytest.mark.parametrize("total_sec, SSV, perturb_sigma, seed", [(30, 10, 0.1, 0)])
    @pytest.mark.parametrize("jit_script", [False, True])
    def test_offline_speaker_clustering(self, n_spks, total_sec, SSV, perturb_sigma, seed, jit_script, cuda=True):
        spk_dur = total_sec / n_spks
        em, ts, mc, mw, spk_ts, gt = generate_toy_data(
            n_spks=n_spks, spk_dur=spk_dur, perturb_sigma=perturb_sigma, torch_seed=seed
        )
        offline_speaker_clustering = SpeakerClustering(maj_vote_spk_count=False, cuda=cuda)
        assert isinstance(offline_speaker_clustering, SpeakerClustering)
        if jit_script:
            offline_speaker_clustering = torch.jit.script(offline_speaker_clustering)

        Y_out = offline_speaker_clustering.forward_infer(
            embeddings_in_scales=em,
            timestamps_in_scales=ts,
            multiscale_segment_counts=mc,
            multiscale_weights=mw,
            oracle_num_speakers=-1,
            max_num_speakers=8,
            enhanced_count_thres=40,
            sparse_search_volume=SSV,
            max_rp_threshold=0.15,
            fixed_thres=-1.0,
        )
        permuted_Y = stitch_cluster_labels(Y_old=gt, Y_new=Y_out)
        permuted_Y = permuted_Y.to(gt.device)
        # mc[-1] is the number of base scale segments
        assert len(set(permuted_Y.tolist())) == n_spks
        assert Y_out.shape[0] == mc[-1]
        assert all(permuted_Y == gt)

    @pytest.mark.run_only_on('CPU')
    @pytest.mark.unit
    @pytest.mark.parametrize("n_spks", [1, 2, 3, 4, 5, 6, 7])
    @pytest.mark.parametrize("total_sec, SSV, perturb_sigma, seed", [(30, 10, 0.1, 0)])
    @pytest.mark.parametrize("jit_script", [False, True])
    def test_offline_speaker_clustering_cpu(self, n_spks, total_sec, SSV, perturb_sigma, seed, jit_script, cuda=False):
        self.test_offline_speaker_clustering(n_spks, total_sec, SSV, perturb_sigma, seed, jit_script, cuda=cuda)

    @pytest.mark.run_only_on('CPU')
    @pytest.mark.unit
    @pytest.mark.parametrize("n_spks", [1])
    @pytest.mark.parametrize("spk_dur", [0.25, 0.5, 0.75, 1, 1.5, 2])
    @pytest.mark.parametrize("SSV, enhanced_count_thres, min_samples_for_nmesc", [(5, 40, 6)])
    @pytest.mark.parametrize("seed", [0])
    def test_offline_speaker_clustering_very_short_cpu(
        self,
        n_spks,
        spk_dur,
        SSV,
        enhanced_count_thres,
        min_samples_for_nmesc,
        seed,
    ):
        em, ts, mc, mw, spk_ts, gt = generate_toy_data(
            n_spks=n_spks, spk_dur=spk_dur, perturb_sigma=0.1, torch_seed=seed
        )
        offline_speaker_clustering = SpeakerClustering(maj_vote_spk_count=False, min_samples_for_nmesc=0, cuda=False)
        assert isinstance(offline_speaker_clustering, SpeakerClustering)
        Y_out = offline_speaker_clustering.forward_infer(
            embeddings_in_scales=em,
            timestamps_in_scales=ts,
            multiscale_segment_counts=mc,
            multiscale_weights=mw,
            oracle_num_speakers=-1,
            max_num_speakers=8,
            enhanced_count_thres=enhanced_count_thres,
            sparse_search_volume=SSV,
            max_rp_threshold=0.15,
            fixed_thres=-1.0,
        )
        permuted_Y = stitch_cluster_labels(Y_old=gt, Y_new=Y_out)
        permuted_Y = permuted_Y.to(gt.device)
        # mc[-1] is the number of base scale segments
        assert len(set(permuted_Y.tolist())) == n_spks
        assert Y_out.shape[0] == mc[-1]
        assert all(permuted_Y == gt)

    @pytest.mark.run_only_on('GPU')
    @pytest.mark.unit
    @pytest.mark.parametrize("spk_dur", [0.25, 0.5, 0.75, 1, 2, 4])
    @pytest.mark.parametrize("n_spks, SSV, enhanced_count_thres, min_samples_for_nmesc", [(1, 5, 40, 6)])
    @pytest.mark.parametrize("seed", [0])
    def test_offline_speaker_clustering_very_short_gpu(
        self,
        n_spks,
        spk_dur,
        SSV,
        enhanced_count_thres,
        min_samples_for_nmesc,
        seed,
    ):
        em, ts, mc, mw, spk_ts, gt = generate_toy_data(
            n_spks=n_spks, spk_dur=spk_dur, perturb_sigma=0.1, torch_seed=seed
        )
        offline_speaker_clustering = SpeakerClustering(maj_vote_spk_count=False, min_samples_for_nmesc=0, cuda=True)
        assert isinstance(offline_speaker_clustering, SpeakerClustering)
        Y_out = offline_speaker_clustering.forward_infer(
            embeddings_in_scales=em,
            timestamps_in_scales=ts,
            multiscale_segment_counts=mc,
            multiscale_weights=mw,
            oracle_num_speakers=-1,
            max_num_speakers=8,
            enhanced_count_thres=enhanced_count_thres,
            sparse_search_volume=SSV,
            max_rp_threshold=0.15,
            fixed_thres=-1.0,
        )
        permuted_Y = stitch_cluster_labels(Y_old=gt, Y_new=Y_out)
        permuted_Y = permuted_Y.to(gt.device)
        # mc[-1] is the number of base scale segments
        assert Y_out.shape[0] == mc[-1]
        assert all(permuted_Y == gt)

    @pytest.mark.run_only_on('CPU')
    @pytest.mark.unit
    @pytest.mark.parametrize("n_spks, SSV, enhanced_count_thres, min_samples_for_nmesc", [(2, 5, 40, 6)])
    @pytest.mark.parametrize("spk_dur, chunk_cluster_count, embeddings_per_chunk", [(120, 4, 50), (240, 4, 100)])
    @pytest.mark.parametrize("seed", [0])
    @pytest.mark.parametrize("jit_script", [False, True])
    def test_longform_speaker_clustering_cpu(
        self,
        n_spks,
        spk_dur,
        SSV,
        enhanced_count_thres,
        min_samples_for_nmesc,
        chunk_cluster_count,
        embeddings_per_chunk,
        jit_script,
        seed,
    ):
        em, ts, mc, mw, spk_ts, gt = generate_toy_data(
            n_spks=n_spks, spk_dur=spk_dur, perturb_sigma=0.1, torch_seed=seed
        )
        longform_speaker_clustering = LongFormSpeakerClustering(cuda=False)
        if jit_script:
            longform_speaker_clustering = torch.jit.script(longform_speaker_clustering)
        else:
            assert isinstance(longform_speaker_clustering, LongFormSpeakerClustering)
        Y_out = longform_speaker_clustering.forward_infer(
            embeddings_in_scales=em,
            timestamps_in_scales=ts,
            multiscale_segment_counts=mc,
            multiscale_weights=mw,
            oracle_num_speakers=-1,
            max_num_speakers=n_spks,
            enhanced_count_thres=enhanced_count_thres,
            sparse_search_volume=SSV,
            max_rp_threshold=0.15,
            fixed_thres=-1.0,
            chunk_cluster_count=chunk_cluster_count,
            embeddings_per_chunk=embeddings_per_chunk,
        )
        permuted_Y = stitch_cluster_labels(Y_old=gt, Y_new=Y_out)
        permuted_Y = permuted_Y.to(gt.device)

        # mc[-1] is the number of base scale segments
        assert Y_out.shape[0] == mc[-1]
        assert all(permuted_Y == gt)

    @pytest.mark.run_only_on('GPU')
    @pytest.mark.unit
    @pytest.mark.parametrize("n_spks, SSV, enhanced_count_thres, min_samples_for_nmesc", [(2, 5, 40, 6)])
    @pytest.mark.parametrize("spk_dur, chunk_cluster_count, embeddings_per_chunk", [(120, 4, 50), (240, 4, 100)])
    @pytest.mark.parametrize("seed", [0])
    @pytest.mark.parametrize("jit_script", [False, True])
    def test_longform_speaker_clustering_gpu(
        self,
        n_spks,
        spk_dur,
        SSV,
        enhanced_count_thres,
        min_samples_for_nmesc,
        chunk_cluster_count,
        embeddings_per_chunk,
        jit_script,
        seed,
    ):
        em, ts, mc, mw, spk_ts, gt = generate_toy_data(
            n_spks=n_spks, spk_dur=spk_dur, perturb_sigma=0.1, torch_seed=seed
        )
        longform_speaker_clustering = LongFormSpeakerClustering(cuda=True)

        if jit_script:
            longform_speaker_clustering = torch.jit.script(longform_speaker_clustering)
        else:
            assert isinstance(longform_speaker_clustering, LongFormSpeakerClustering)

        Y_out = longform_speaker_clustering.forward_infer(
            embeddings_in_scales=em,
            timestamps_in_scales=ts,
            multiscale_segment_counts=mc,
            multiscale_weights=mw,
            oracle_num_speakers=-1,
            max_num_speakers=n_spks,
            enhanced_count_thres=enhanced_count_thres,
            sparse_search_volume=SSV,
            max_rp_threshold=0.15,
            fixed_thres=-1.0,
            chunk_cluster_count=chunk_cluster_count,
            embeddings_per_chunk=embeddings_per_chunk,
        )
        permuted_Y = stitch_cluster_labels(Y_old=gt, Y_new=Y_out)
        permuted_Y = permuted_Y.to(gt.device)

        # mc[-1] is the number of base scale segments
        assert Y_out.shape[0] == mc[-1]
        assert all(permuted_Y == gt)

    @pytest.mark.run_only_on('GPU')
    @pytest.mark.unit
    @pytest.mark.parametrize("n_spks", [1, 2, 3])
    @pytest.mark.parametrize("total_sec, buffer_size, sigma", [(30, 30, 0.1)])
    @pytest.mark.parametrize("seed", [0])
    @pytest.mark.parametrize("jit_script", [False, True])
    def test_online_speaker_clustering(self, n_spks, total_sec, buffer_size, sigma, seed, jit_script, cuda=True):
        step_per_frame = 2
        spk_dur = total_sec / n_spks
        em, ts, mc, _, _, gt = generate_toy_data(n_spks, spk_dur=spk_dur, perturb_sigma=sigma, torch_seed=seed)
        em_s, ts_s = split_input_data(em, ts, mc)

        emb_gen = em_s[-1]
        segment_indexes = ts_s[-1]
        if cuda:
            device = torch.cuda.current_device()
            emb_gen, segment_indexes = emb_gen.to(device), segment_indexes.to(device)

        history_buffer_size = buffer_size
        current_buffer_size = buffer_size

        online_clus = OnlineSpeakerClustering(
            max_num_speakers=8,
            max_rp_threshold=0.15,
            sparse_search_volume=30,
            history_buffer_size=history_buffer_size,
            current_buffer_size=current_buffer_size,
            cuda=cuda,
        )
        if jit_script:
            online_clus = torch.jit.script(online_clus)

        n_frames = int(emb_gen.shape[0] / step_per_frame)
        evaluation_list = []

        # Simulate online speaker clustering
        for frame_index in range(n_frames):
            curr_emb = emb_gen[0 : (frame_index + 1) * step_per_frame]
            base_segment_indexes = torch.arange(curr_emb.shape[0]).to(curr_emb.device)
            # Check history_buffer_size and history labels
            assert (
                online_clus.history_embedding_buffer_emb.shape[0] <= history_buffer_size
            ), "History buffer size error"
            assert (
                online_clus.history_embedding_buffer_emb.shape[0]
                == online_clus.history_embedding_buffer_label.shape[0]
            )

            # Call clustering function
            merged_clus_labels = online_clus.forward_infer(
                curr_emb=curr_emb, base_segment_indexes=base_segment_indexes, frame_index=frame_index, cuda=cuda
            )

            # Resolve permutations
            assert len(merged_clus_labels) == (frame_index + 1) * step_per_frame
            # Resolve permutation issue by using stitch_cluster_labels function
            merged_clus_labels = merged_clus_labels.cpu()
            merged_clus_labels = stitch_cluster_labels(Y_old=gt[: len(merged_clus_labels)], Y_new=merged_clus_labels)
            evaluation_list.extend(list(merged_clus_labels == gt[: len(merged_clus_labels)]))

        assert online_clus.is_online
        cumul_label_acc = sum(evaluation_list) / len(evaluation_list)
        assert cumul_label_acc > 0.9

    @pytest.mark.run_only_on('CPU')
    @pytest.mark.unit
    @pytest.mark.parametrize("n_spks, total_sec, buffer_size, sigma, seed", [(3, 30, 30, 0.1, 0)])
    @pytest.mark.parametrize("jit_script", [False, True])
    def test_online_speaker_clustering_cpu(self, n_spks, total_sec, buffer_size, sigma, seed, jit_script, cuda=False):
        self.test_online_speaker_clustering(n_spks, total_sec, buffer_size, sigma, seed, jit_script, cuda)


class TestLinearSumAssignmentAlgorithm:
    @pytest.mark.unit
    def test_lsa_solver_export_test(self):
        cost_matrix = torch.randint(0, 10, (3, 3))
        solver = LinearSumAssignmentSolver(cost_matrix)
        solver = torch.jit.script(solver)
        assert isinstance(solver, torch.jit._script.RecursiveScriptClass)

    @pytest.mark.unit
    @pytest.mark.parametrize(
        "cost_matrix",
        [torch.tensor([[7, 6, 2, 9, 2], [6, 2, 1, 3, 9], [5, 6, 8, 9, 5], [6, 8, 5, 8, 6], [9, 5, 6, 4, 7]])],
    )
    def test_linear_sum_assignment_algorithm_cost_matrix(self, cost_matrix):
        """
        Test the linear sum assignment algorithm with a cost matrix

        Compare with the scipy implementation and make sure the final cost is the same.
        NOTE: There could be multiple solutions with the same cost in linear sum assignment problem.
              This test only checks if the cost is the same.
        """
        row_ind_nm, col_ind_nm = nemo_linear_sum_assignment(cost_matrix)
        row_ind_sc, col_ind_sc = scipy_linear_sum_assignment(cost_matrix.cpu().numpy())
        cost_nm = sum(cost_matrix[row_ind_nm, col_ind_nm])
        cost_sc = sum(cost_matrix[row_ind_sc, col_ind_sc])
        assert cost_nm == cost_sc

    @pytest.mark.unit
    @pytest.mark.parametrize("seed", [0, 1])
    @pytest.mark.parametrize("mat_size", [1, 2, 4, 8])
    def test_linear_sum_assignment_algorithm_random_matrix(self, seed, mat_size):
        torch.manual_seed(seed)
        cost_matrix = torch.randint(0, 10, (mat_size, mat_size))
        self.test_linear_sum_assignment_algorithm_cost_matrix(cost_matrix)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import defaultdict

import sacrebleu
import torch
from whisper_normalizer.english import EnglishTextNormalizer

from nemo.collections.asr.models import ASRModel
from nemo.collections.common.parts.optional_cuda_graphs import WithOptionalCudaGraphs
from nemo.collections.speechlm2.parts.precision import fp32_precision
from nemo.collections.speechlm2.parts.pretrained import load_pretrained_nemo
from nemo.utils import logging


class ASRBLEU:
    """
    Computes BLEU scores on ASR predictions on generated audio with pretrained NeMo ASR.
    By default, uses Whisper's EnglishTextNormalizer on hypotheses and references.
    """

    def __init__(self, pretrained_asr: str, normalize: bool = True, normalizer=None, verbose: bool = True) -> None:
        self.asr = None  # load into memory on reset()
        self.pretrained_asr_name = pretrained_asr
        self.verbose = verbose
        if normalize:
            if normalizer is None:
                self.normalizer = EnglishTextNormalizer()
            else:
                self.normalizer = normalizer
        else:
            self.normalizer = _identity

        self._refs = defaultdict(list)
        self._hyps = defaultdict(list)

    def reset(self):
        # Cleaning up GPU memory before we load ASRModel, because it may already
        # be quite fragmented and close to the limit after observing many
        # dynamic shapes during the training epoch.
        torch.cuda.memory.empty_cache()
        with fp32_precision():  # Some NeMo ASR models weren't trained with bfloat16.
            self.asr = load_pretrained_nemo(ASRModel, self.pretrained_asr_name).eval()
        WithOptionalCudaGraphs.disable_cuda_graphs_recursive(self.asr, attribute_path="decoding.decoding")
        return self

    def update(
        self, name: str, refs: list[str], pred_audio: torch.Tensor, pred_audio_lens: torch.Tensor = None
    ) -> list[str]:
        if self.asr is None:
            self.reset()

        if pred_audio_lens is None:
            pred_audio_lens = [pred_audio.shape[1]] * pred_audio.shape[0]

        with fp32_precision():
            asr_hyps = self.asr.transcribe(
                [audio[:alen] for audio, alen in zip(pred_audio, pred_audio_lens)],
                batch_size=pred_audio.shape[0],
                verbose=False,
            )
        asr_hyps_texts = []
        for ref, asr_hyp in zip(refs, asr_hyps):
            asr_hyp = asr_hyp.text
            self._refs[name].append(self.normalizer(ref))
            self._hyps[name].append(self.normalizer(asr_hyp))
            if self.verbose:
                asrb = sacrebleu.sentence_bleu(asr_hyp, [ref]).score
                logging.info(f"[REF]\t{ref}\n[ASR]\t{asr_hyp} [{asrb:.2f}]")
            asr_hyps_texts.append(asr_hyp)

        return asr_hyps_texts

    def compute(self) -> dict[str, torch.Tensor]:
        """Computes the final score and deallocates ASR and partial results."""
        corpus_metric = {}
        for name in self._refs.keys():
            metric = torch.tensor(sacrebleu.corpus_bleu(self._hyps[name], [self._refs[name]]).score)
            corpus_metric[f"asr_bleu_{name}"] = metric
        corpus_metric["asr_bleu"] = torch.stack(list(corpus_metric.values())).mean()
        self._refs.clear()
        self._hyps.clear()
        self.asr = None  # free up GPU memory
        torch.cuda.memory.empty_cache()
        return corpus_metric


def _identity(x):
    return x



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=C0115,C0116
from dataclasses import dataclass, field
from typing import Dict, List

import torch
import torch.nn.functional as F
from megatron.energon import VQASample, batch_list, batch_pad_stack
from torch.nn.utils.rnn import pad_sequence

from nemo.collections.multimodal.data.energon.sample_encoder import SampleEncoder
from nemo.collections.multimodal.data.energon.task_encoder import MultiModalTaskEncoder
from nemo.collections.vlm.mllama.data.sample_encoder import Llama3SampleEncoder, LlamaImageTextSample


def pad_or_truncate(sequence_batch, seq_length: int, padding_value: int):
    # Pad the sequence if it's shorter than seq_length
    if sequence_batch.size(1) < seq_length:
        pad_size = seq_length - sequence_batch.size(1)
        sequence_batch = F.pad(sequence_batch, (0, pad_size), value=padding_value)
    else:
        # Truncate the sequence if it's longer than seq_length
        sequence_batch = sequence_batch[:, :seq_length]

    return sequence_batch


@dataclass
class LlamaImageTextRawBatch:
    __keys__: List[str] = field(default_factory=list)

    tokens: torch.Tensor = field(default_factory=lambda: torch.empty(0, dtype=torch.long))
    labels: torch.Tensor = field(default_factory=lambda: torch.empty(0, dtype=torch.long))
    loss_mask: torch.Tensor = field(default_factory=lambda: torch.empty(0, dtype=torch.float))

    batch_images: torch.Tensor = field(default_factory=lambda: torch.empty(0))
    batch_masks: torch.Tensor = field(default_factory=lambda: torch.empty(0))

    aspect_ratio_ids: torch.Tensor = field(default_factory=lambda: torch.empty(0, dtype=torch.float))
    aspect_ratio_mask: torch.Tensor = field(default_factory=lambda: torch.empty(0, dtype=torch.float))
    num_chunks: torch.Tensor = field(default_factory=lambda: torch.empty(0, dtype=torch.float))


class LlamaTaskEncoder(MultiModalTaskEncoder):
    def __init__(self, tokenizer, image_processor, multimodal_sample_config, seq_length=None):
        super().__init__(tokenizer, image_processor, multimodal_sample_config)
        self.encoders: Dict[str, SampleEncoder] = {
            VQASample.__name__: Llama3SampleEncoder(tokenizer, image_processor, multimodal_sample_config)
        }
        self.seq_length = seq_length
        self.ignore_index = multimodal_sample_config.ignore_place_holder

    def batch(self, samples: List[LlamaImageTextSample]) -> LlamaImageTextRawBatch:

        keys, images, tokens, labels, loss_mask, vision_mask = [], [], [], [], [], []
        aspect_ratio_ids, aspect_ratio_mask, num_tiles = [], [], []
        for sample in samples:
            keys.append(sample.__key__)
            images.append(sample.images)
            tokens.append(sample.tokens)
            labels.append(sample.labels)
            loss_mask.append(sample.loss_mask)
            vision_mask.append(sample.vision_mask)
            aspect_ratio_ids.append(sample.aspect_ratio_ids)
            aspect_ratio_mask.append(sample.aspect_ratio_mask)
            num_tiles.append(sample.num_tiles)

        batch_keys = batch_list(keys)
        batch_images = batch_pad_stack(images)

        batch_tokens = pad_sequence(tokens, batch_first=True)
        batch_labels = pad_sequence(labels, batch_first=True, padding_value=self.ignore_index)
        batch_loss_mask = batch_pad_stack(loss_mask)
        if self.seq_length is not None:
            seq_length = self.seq_length
        else:
            seq_length = (batch_tokens.size(1) - 1) // 64 * 64 + 64
        batch_tokens = pad_or_truncate(batch_tokens, seq_length, 0)
        batch_labels = pad_or_truncate(batch_labels, seq_length, self.ignore_index)
        batch_loss_mask = pad_or_truncate(batch_loss_mask, seq_length, 0)
        assert batch_loss_mask.sum() > 0, "This batch has nothing to predict! Will trigger a nan loss."
        batch_vision_mask = batch_pad_stack(vision_mask)
        batch_aspect_ratio_ids = batch_pad_stack(aspect_ratio_ids)
        batch_aspect_ratio_mask = batch_pad_stack(aspect_ratio_mask)
        batch_num_tiles = torch.tensor(num_tiles)
        return LlamaImageTextRawBatch(
            __keys__=batch_keys,
            batch_images=batch_images,
            batch_masks=batch_vision_mask,
            tokens=batch_tokens,
            labels=batch_labels,
            loss_mask=batch_loss_mask,
            aspect_ratio_ids=batch_aspect_ratio_ids,
            aspect_ratio_mask=batch_aspect_ratio_mask,
            num_chunks=batch_num_tiles,
        )



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import re
from dataclasses import dataclass
from typing import Callable, Optional, Union

import torch.distributed
import torch.nn.functional as F
from megatron.core.models.vision.clip_vit_model import CLIPViTModel as MCoreCLIPViTModel
from megatron.core.models.vision.multimodal_projector import MultimodalProjector as MCoreMultimodalProjector
from megatron.core.tensor_parallel.layers import ColumnParallelLinear

try:
    from megatron.core.transformer.custom_layers.transformer_engine import (
        TEColumnParallelLinear,
        TENorm,
        TERowParallelLinear,
    )
except ImportError:
    from nemo.utils import logging

    # These Defaults are needed to make sure the code compiles
    TEColumnParallelLinear = None
    TENorm = None
    TERowParallelLinear = None
    logging.warning(
        "Failed to import Transformer Engine dependencies. "
        "`from megatron.core.transformer.custom_layers.transformer_engine import *`"
        "If using NeMo Run, this is expected. Otherwise, please verify the Transformer Engine installation."
    )

from megatron.core.transformer.mlp import MLP, MLPSubmodules
from megatron.core.transformer.spec_utils import ModuleSpec
from megatron.core.transformer.transformer_config import TransformerConfig
from transformers import CLIPVisionConfig, CLIPVisionModel

from nemo.lightning import io


def set_input_tensor(self, tensor):
    """Sets input tensor func place holder"""
    pass


def get_image_sequence_length(img_h, img_w, patch_dim, add_class_token, class_token_len):
    """Get image sequence length given image size, patch size, and class token."""
    num_patches_per_dim_h = img_h // patch_dim
    num_patches_per_dim_w = img_w // patch_dim
    num_patches = num_patches_per_dim_h * num_patches_per_dim_w
    return num_patches + (class_token_len if add_class_token else 0)


class DownSampleBlock(torch.nn.Module):
    """Downsample block following the ViLA-VLM paper."""

    # pylint: disable=line-too-long
    # Implement from https://github.com/NVlabs/VILA/blob/3522eef015e48d73cf83fc2b949cd464dab1ba3c/llava/model/multimodal_projector/base_projector.py#L48
    # small adjusmtnet with x.transpose(0, 1)

    def forward(self, x):
        """Downsample the input tensor."""
        x = x.transpose(0, 1)
        vit_embeds = x
        h = w = int(vit_embeds.shape[1] ** 0.5)
        vit_embeds = vit_embeds.reshape(vit_embeds.shape[0], h, w, -1)
        vit_embeds = self.flat_square(vit_embeds)
        vit_embeds = vit_embeds.reshape(vit_embeds.shape[0], -1, vit_embeds.shape[-1])
        return vit_embeds

    def flat_square(self, x):
        """Flatten the input tensor and make it square."""
        n, w, h, c = x.size()
        if w % 2 == 1:
            x = torch.concat([x, torch.zeros((n, 1, h, c), dtype=x.dtype).to(x.device)], dim=1).contiguous()
            n, w, h, c = x.size()
        if h % 2 == 1:
            x = torch.concat([x, torch.zeros((n, w, 1, c), dtype=x.dtype).to(x.device)], dim=2).contiguous()
            n, w, h, c = x.size()
        x = x.contiguous()
        x = x.view(n, w, int(h / 2), int(c * 2))
        x = x.permute(0, 2, 1, 3).contiguous()
        x = x.view(n, int(h / 2), int(w / 2), int(c * 4))
        x = x.permute(0, 2, 1, 3).contiguous()
        return x


@dataclass
class MultimodalProjectorConfig(TransformerConfig, io.IOMixin):
    """
    For MLP, fc1 in shape of input_size, ffn_hidden_size, fc2 in shape of ffn_hidden_size, hidden_size
    """

    projector_type: str = "mlp2x_gelu"
    layer_spec: Optional[MLPSubmodules] = None
    input_size: Optional[int] = 1024
    hidden_size: int = 1024
    ffn_hidden_size: int = 1024
    activation_func: Callable = F.gelu
    bias: bool = True
    bias_activation_fusion: bool = True
    num_layers: int = 1  # placeholder, NOT used!
    num_attention_heads: int = 8  # placeholder, NOT used!

    def configure_model(self) -> "MCoreMultimodalProjector":
        # pylint: disable=C0115,C0116
        if self.projector_type.startswith("mcore") and self.layer_spec is None:
            self.add_bias_linear = self.bias
            if self.projector_type == "mcore_mlp":
                self.projector_type = "mlp"  # strip "mcore_" for mcore init
                self.layer_spec = ModuleSpec(
                    module=MLP,
                    submodules=MLPSubmodules(
                        linear_fc1=TEColumnParallelLinear,
                        linear_fc2=TERowParallelLinear,
                    ),
                )
                self.layer_spec = self.layer_spec.submodules
            elif self.projector_type == "mcore_affine":
                self.projector_type = "affine"  # strip "mcore_" for mcore init
                self.layer_spec = MLPSubmodules(linear_fc1=ColumnParallelLinear, linear_fc2=None)
            else:
                raise NotImplementedError(f"Not supported projector type `{self.projector_type}`")

            return MCoreMultimodalProjector(
                self,
                self.layer_spec,
                projector_type=self.projector_type,
                input_size=self.input_size,
            )

        # if using vila's downsample + mlp projector
        if self.projector_type == "vila_downsample_mlp":
            model = torch.nn.Sequential(
                DownSampleBlock(),
                torch.nn.LayerNorm(self.input_size * 4, dtype=self.params_dtype),
                torch.nn.Linear(self.input_size * 4, self.hidden_size, bias=True, dtype=self.params_dtype),
                torch.nn.GELU(),
                torch.nn.Linear(self.hidden_size, self.hidden_size, bias=True, dtype=self.params_dtype),
            )
            from types import MethodType

            model.set_input_tensor = MethodType(set_input_tensor, model)
            return model

        # e.g. "mlp2x_gelu"
        mlp_gelu_match = re.match(r'^mlp(\d+)x_gelu$', self.projector_type)
        if mlp_gelu_match:
            mlp_depth = int(mlp_gelu_match.group(1))
            modules = [torch.nn.Linear(self.input_size, self.ffn_hidden_size, bias=True, dtype=self.params_dtype)]
            for _ in range(1, mlp_depth):
                modules.append(torch.nn.GELU())
                modules.append(
                    torch.nn.Linear(self.ffn_hidden_size, self.hidden_size, bias=True, dtype=self.params_dtype)
                )
            model = torch.nn.Sequential(*modules)
            from types import MethodType

            model.set_input_tensor = MethodType(set_input_tensor, model)
        else:
            raise NotImplementedError(f"Not supported projector type `{self.projector_type}`")

        return model


@dataclass
class HFCLIPVisionConfig(CLIPVisionConfig, io.IOMixin):
    """
    https://github.com/huggingface/transformers/blob/v4.44.0/src/transformers/models/clip/configuration_clip.py#L261
    """

    hidden_size: int = 1024
    add_class_token: bool = False
    class_token_len: int = 1
    num_image_embeddings_per_tile: Optional[int] = None
    pretrained_model_name_or_path: Optional[Union[str, os.PathLike]] = None

    def __post_init__(self, *args, **kwargs) -> None:
        # pylint: disable=C0115,C0116
        CLIPVisionConfig.__init__(self, *args, **kwargs, hidden_size=self.hidden_size)
        if self.pretrained_model_name_or_path is not None:
            config = CLIPVisionConfig.from_pretrained(self.pretrained_model_name_or_path)
            for key, value in config.to_dict().items():
                setattr(self, key, value)
        self.num_image_embeddings_per_tile = get_image_sequence_length(
            img_h=self.image_size,
            img_w=self.image_size,
            patch_dim=self.patch_size,
            add_class_token=self.add_class_token,
            class_token_len=self.class_token_len,
        )

    def configure_model(self) -> "CLIPVisionModel":
        # pylint: disable=C0115,C0116
        # Monkey patch the method to the vision encoder
        CLIPVisionModel.set_input_tensor = set_input_tensor

        if self.pretrained_model_name_or_path is None:
            model = CLIPVisionModel(self)
        else:
            model = CLIPVisionModel.from_pretrained(self.pretrained_model_name_or_path)

        # add attribute "tensor_parallel_grad_reduce" to the model for TP grad all-reduce
        model.tensor_parallel_grad_reduce = True

        return model


@dataclass
class CLIPViTConfig(TransformerConfig, io.IOMixin):
    """MCore CLIP ViT Config"""

    ln_pre_impl: Union[ModuleSpec, type] = TENorm
    ln_post_impl: Union[ModuleSpec, type] = TENorm
    add_class_token: bool = True
    class_token_len: int = 1
    patch_dim: int = 14
    img_h: int = 336
    img_w: int = 336
    vision_model_type: str = "clip"  # ["clip", "siglip", "internvit"]
    num_image_embeddings_per_tile: Optional[int] = None
    transformer_layer_spec: ModuleSpec = None

    num_layers: int = 1  # Placeholder, NOT used!
    num_attention_heads: int = 8  # Placeholder, NOT used!

    def __post_init__(self):
        # pylint: disable=C0115,C0116
        if self.vision_model_type == "siglip":
            self.add_class_token = False
            self.class_token_len = 0
        self.num_image_embeddings_per_tile = get_image_sequence_length(
            img_h=self.img_h,
            img_w=self.img_w,
            patch_dim=self.patch_dim,
            add_class_token=self.add_class_token,
            class_token_len=self.class_token_len,
        )

    def configure_model(self) -> "BaseCLIPViTModel":
        # pylint: disable=C0115,C0116
        transformer_layer_spec = self.transformer_layer_spec
        if not isinstance(transformer_layer_spec, ModuleSpec):
            from nemo.collections.vlm.layer_specs import get_layer_spec_te

            transformer_layer_spec = get_layer_spec_te(is_vit=True)
        return BaseCLIPViTModel(
            self,
            transformer_layer_spec,
            ln_pre_impl=self.ln_pre_impl,
            ln_post_impl=self.ln_post_impl,
            add_class_token=self.add_class_token,
            class_token_len=self.class_token_len,
            patch_dim=self.patch_dim,
            img_h=self.img_h,
            img_w=self.img_w,
            model_subtype=self.vision_model_type,
        )


class BaseCLIPViTModel(MCoreCLIPViTModel):
    """CLIP ViT vision model."""

    def forward(
        self, x: torch.Tensor, attention_mask: Optional[torch.Tensor] = None, num_unused_layers: int = 0
    ) -> torch.Tensor:
        # pylint: disable=C0115,C0116
        if num_unused_layers > 0:
            unused_layers = self.decoder.layers[-num_unused_layers:]
            self.decoder.layers = self.decoder.layers[:-num_unused_layers]
            x = super().forward(x, attention_mask)
            self.decoder.layers.extend(unused_layers)
            return x

        return super().forward(x, attention_mask)



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import time
from typing import Dict, List

import torch
import torch.profiler
from diffusers import AutoencoderKL
from torch import nn


class VAEGenerator:
    """
    A class for generating and searching different Variational Autoencoder (VAE) configurations.

    This class provides functionality to generate various VAE architecture configurations
    given a specific input resolution and compression ratio. It allows searching through a
    design space to find configurations that match given parameter and memory budgets.
    """

    def __init__(self, input_resolution: int = 1024, compression_ratio: int = 16) -> None:
        if input_resolution == 1024:
            assert compression_ratio in [8, 16]
        elif input_resolution == 2048:
            assert compression_ratio in [8, 16, 32]
        else:
            raise NotImplementedError("Higher resolution than 2028 is not implemented yet!")

        self._input_resolution = input_resolution
        self._compression_ratio = compression_ratio

    def _generate_input(self):
        """
        Generate a random input tensor with the specified input resolution.

        The tensor is placed on the GPU in half-precision (float16).
        """
        random_tensor = torch.rand(1, 3, self.input_resolution, self.input_resolution)
        random_tensor = random_tensor.to(dtype=torch.float16, device="cuda")
        return random_tensor

    def _count_parameters(self, model: nn.Module = None):
        """
        Count the number of trainable parameters in a given model.

        Args:
            model (nn.Module): The model for which to count parameters.

        Returns:
            int: The number of trainable parameters.
        """
        assert model is not None, "Please provide a nn.Module to count the parameters."
        return sum(p.numel() for p in model.parameters() if p.requires_grad)

    def _load_base_json_skeleton(self):
        """
        Load a base configuration skeleton for the VAE.

        Returns:
            dict: A dictionary representing the base configuration JSON skeleton.
        """
        skeleton = {
            "_class_name": "AutoencoderKL",
            "_diffusers_version": "0.20.0.dev0",
            "_name_or_path": "../sdxl-vae/",
            "act_fn": "silu",
            "block_out_channels": [],
            "down_block_types": [],
            "force_upcast": False,
            "in_channels": 3,
            "latent_channels": -1,  # 16
            "layers_per_block": -1,  # 2
            "norm_num_groups": 32,
            "out_channels": 3,
            "sample_size": 1024,  # resolution size
            "scaling_factor": 0.13025,
            "up_block_types": [],
        }
        return skeleton

    def _generate_all_combinations(self, attr):
        """
        Generates all possible combinations from a search space dictionary.

        Args:
            attr (dict): A dictionary where each key has a list of possible values.

        Returns:
            List[Dict]: A list of dictionaries, each representing a unique combination of attributes.
        """
        keys = list(attr.keys())
        choices = [attr[key] for key in keys]
        all_combinations = list(itertools.product(*choices))

        combination_dicts = []
        for combination in all_combinations:
            combination_dict = {key: value for key, value in zip(keys, combination)}
            combination_dicts.append(combination_dict)

        return combination_dicts

    def _assign_attributes(self, choice):
        """
        Assign a chosen set of attributes to the base VAE configuration skeleton.

        Args:
            choice (dict): A dictionary of attributes to assign to the skeleton.

        Returns:
            dict: A dictionary representing the updated VAE configuration.
        """
        search_space_skleton = self._load_base_json_skeleton()
        search_space_skleton["down_block_types"] = choice["down_block_types"]
        search_space_skleton["up_block_types"] = choice["up_block_types"]
        search_space_skleton["block_out_channels"] = choice["block_out_channels"]
        search_space_skleton["layers_per_block"] = choice["layers_per_block"]
        search_space_skleton["latent_channels"] = choice["latent_channels"]
        return search_space_skleton

    def _search_space_16x1024(self):
        """
        Define the search space for a 16x compression ratio at 1024 resolution.

        Returns:
            dict: A dictionary defining lists of possible attribute values.
        """
        attr = {}
        attr["down_block_types"] = [["DownEncoderBlock2D"] * 5]
        attr["up_block_types"] = [["UpDecoderBlock2D"] * 5]
        attr["block_out_channels"] = [
            [128, 256, 512, 512, 512],
            [128, 256, 512, 512, 1024],
            [128, 256, 512, 1024, 2048],
            [64, 128, 256, 512, 512],
        ]
        attr["layers_per_block"] = [1, 2, 3]
        attr["latent_channels"] = [4, 16, 32, 64]
        return attr

    def _search_space_8x1024(self):
        """
        Define the search space for an 8x compression ratio at 1024 resolution.

        Returns:
            dict: A dictionary defining lists of possible attribute values.
        """
        attr = {}
        attr["down_block_types"] = [["DownEncoderBlock2D"] * 4]
        attr["up_block_types"] = [["UpDecoderBlock2D"] * 4]
        attr["block_out_channels"] = [[128, 256, 512, 512], [128, 256, 512, 1024], [64, 128, 256, 512]]
        attr["layers_per_block"] = [1, 2, 3]
        attr["latent_channels"] = [4, 16, 32, 64]
        return attr

    def _sort_data_in_place(self, data: List[Dict], mode: str) -> None:
        """
        Sort the list of design configurations in place based on a chosen mode.

        Args:
            data (List[Dict]): A list of dictionaries representing design configurations.
            mode (str): The sorting criterion. Can be 'abs_param_diff', 'abs_cuda_mem_diff', or 'mse'.
        """
        if mode == 'abs_param_diff':
            data.sort(key=lambda x: abs(x['param_diff']))
        elif mode == 'abs_cuda_mem_diff':
            data.sort(key=lambda x: abs(x['cuda_mem_diff']))
        elif mode == 'mse':
            data.sort(key=lambda x: (x['param_diff'] ** 2 + x['cuda_mem_diff'] ** 2) / 2)
        else:
            raise ValueError("Invalid mode. Choose from 'abs_param_diff', 'abs_cuda_mem_diff', 'mse'.")

    def _print_table(self, data, headers, col_widths):
        """
        Print a formatted table of the design choices.

        Args:
            data (List[Dict]): The data to print, each entry a design configuration.
            headers (List[str]): Column headers.
            col_widths (List[int]): Widths for each column.
        """
        # Create header row
        header_row = ""
        for header, width in zip(headers, col_widths):
            header_row += f"{header:<{width}}"
        print(header_row)
        print("-" * sum(col_widths))

        # Print each data row
        for item in data:
            row = f"{item['param_diff']:<{col_widths[0]}}"
            row += f"{item['cuda_mem_diff']:<{col_widths[1]}}"
            print(row)

    def search_for_target_vae(self, parameters_budget=0, cuda_max_mem=0):
        """
        Search through available VAE design choices to find one that best matches
        the given parameter and memory budgets.

        Args:
            parameters_budget (float, optional): The target number of parameters (in millions).
            cuda_max_mem (float, optional): The target maximum GPU memory usage (in MB).

        Returns:
            AutoencoderKL: The chosen VAE configuration that best matches the provided budgets.
        """
        if parameters_budget <= 0 and cuda_max_mem <= 0:
            raise ValueError("Please specify a valid parameter budget or cuda max memory budget")

        search_space_choices = []
        if self.input_resolution == 1024 and self.compression_ratio == 8:
            search_space = self._search_space_8x1024()
            search_space_choices = self._generate_all_combinations(search_space)
        elif self.input_resolution == 1024 and self.compression_ratio == 16:
            search_space = self._search_space_16x1024()
            search_space_choices = self._generate_all_combinations(search_space)

        inp_tensor = self._generate_input()
        inp_tensor = inp_tensor.to(dtype=torch.float16, device="cuda")
        design_choices = []

        for choice in search_space_choices:
            parameters_budget_diff = 0
            cuda_max_mem_diff = 0

            curt_design_json = self._assign_attributes(choice)
            print("-" * 20)
            print(choice)
            vae = AutoencoderKL.from_config(curt_design_json)
            vae = vae.to(dtype=torch.float16, device="cuda")
            total_params = self._count_parameters(vae)
            total_params /= 10**6
            # Reset peak memory statistics
            torch.cuda.reset_peak_memory_stats()
            torch.cuda.synchronize()

            with torch.profiler.profile(
                activities=[
                    torch.profiler.ProfilerActivity.CPU,
                    torch.profiler.ProfilerActivity.CUDA,
                ],
                profile_memory=True,  # Enables memory profiling
                record_shapes=True,  # Records tensor shapes
                with_stack=True,  # Records stack traces
            ) as prof:
                # Perform forward pass
                start_time = time.perf_counter()
                with torch.no_grad():
                    _ = vae.encode(inp_tensor).latent_dist.sample()
                torch.cuda.synchronize()
                end_time = time.perf_counter()

            total_execution_time_ms = (end_time - start_time) * 1000

            # Get maximum memory allocated
            max_memory_allocated = torch.cuda.max_memory_allocated()
            max_memory_allocated = max_memory_allocated / (1024**2)

            parameters_budget_diff = parameters_budget - total_params
            cuda_max_mem_diff = cuda_max_mem - max_memory_allocated
            design_choices.append(
                {"param_diff": parameters_budget_diff, "cuda_mem_diff": cuda_max_mem_diff, "design": curt_design_json}
            )

            print(f"  Total params: {total_params}")
            print(f"  Max GPU Memory Usage: {max_memory_allocated} MB")
            print(f"  Total Execution Time: {total_execution_time_ms:.2f} ms")

            print(prof.key_averages().table(sort_by="cuda_time_total", row_limit=10))

            print("-" * 20)
        sort_mode = "abs_param_diff"
        if parameters_budget == 0:
            sort_mode = "abs_cuda_mem_diff"
        elif cuda_max_mem == 0:
            sort_mode = "abs_param_diff"
        else:
            sort_mode = "mse"

        print("#" * 20)
        self._sort_data_in_place(design_choices, sort_mode)
        headers = ["param_diff (M)", "cuda_mem_diff (MB)"]
        col_widths = [12, 15]
        self._print_table(design_choices, headers, col_widths)

        vae = AutoencoderKL.from_config(design_choices[0]["design"])
        return vae

    @property
    def input_resolution(self) -> int:
        """
        Get the input resolution for the VAE.

        Returns:
            int: The input resolution.
        """
        return self._input_resolution

    @property
    def compression_ratio(self) -> float:
        """
        Get the compression ratio for the VAE.

        Returns:
            float: The compression ratio.
        """
        return self._compression_ratio



# Copyright (c) 2025, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
try:
    import utmosv2
except ImportError:
    raise ImportError(
        "UTMOSv2 is not installed. Please install it using `pip install git+https://github.com/sarulab-speech/UTMOSv2.git@v1.2.1`."
    )
from typing import Optional
import torch
from threadpoolctl import threadpool_limits

"""
Uses the UTMOSv2 model to estimate the MOS of a speech audio file.
"""


class UTMOSv2Calculator:
    """
    Wrapper around UTMOSv2 MOS estimator to make it easy to use.
    Args:
        device: The device to place the model on. If None, the best available device will be used.
            Default is None.
    """

    def __init__(self, device: Optional[str] = None):
        if device is None:
            device = get_available_device()
        self.model = utmosv2.create_model()
        self.model.eval()
        self.model.to(torch.device(device))

    def __call__(self, file_path):
        """
        Estimate the MOS of the given speech audio file using UTMOSv2.
        """
        with torch.inference_mode():
            # UTMOSv2 tends to launch many OpenMP threads which can overload the machine's CPUs
            # without actually speeding up prediction. Limit to 4 threads.
            with threadpool_limits(limits=4):
                mos_score = self.model.predict(input_path=file_path, num_repetitions=1, num_workers=0)
        return mos_score

    def process_directory(self, input_dir: str, batch_size: int = 16) -> list[dict[str, str | float]]:
        """
        Computes UTMOSv2 scores for all `*.wav` files in the given directory.
        Args:
            input_dir: The directory containing the audio files.
            batch_size: The number of audio files to process in parallel.
        Returns:
            A list of dictionaries, each containing the file path and the UTMOSv2 score.
        """
        with torch.inference_mode():
            # UTMOSV2 tends to launch many of OpenMP threads which overloads the machine's CPUs
            # while actually slowing down the prediction. Limit the number of threads here.
            with threadpool_limits(limits=1):
                results = self.model.predict(
                    input_dir=input_dir, num_repetitions=1, num_workers=batch_size, batch_size=batch_size
                )
        return results


def get_available_device():
    """
    Get the best available device (prefer GPU, fallback to CPU).
    """
    if torch.cuda.is_available():
        return "cuda:0"  # Use first GPU
    else:
        return "cpu"



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest
from lhotse import SupervisionSegment
from lhotse.array import Array, TemporalArray
from lhotse.audio import AudioSource, Recording
from lhotse.cut import MonoCut

from nemo.collections.common.data.lhotse.sampling import (
    CERFilter,
    ContextSpeakerSimilarityFilter,
    ValidationStatusFilter,
)


@pytest.fixture
def cut_example():
    cut = MonoCut(
        id='cut-rec-Zdud2gXLTXY-238.16-6.88_repeat0',
        start=238.16,
        duration=6.88,
        channel=0,
        supervisions=[
            SupervisionSegment(
                id='sup-rec-Zdud2gXLTXY',
                recording_id='rec-Zdud2gXLTXY',
                start=238.16,
                duration=6.88,
                channel=0,
                text='and in like manner, as do other parts in which there appears to exist an adaptation to an end.',
                language='en',
                speaker='| Language:en Dataset:nvyt2505 Speaker:Zdud2gXLTXY_SPEAKER_02 |',
                gender=None,
                custom={
                    'cer': 0.03,
                    'bandwidth': 10875,
                    'stoi_squim': 0.921,
                    'sisdr_squim': 15.17,
                    'pesq_squim': 1.845,
                    'dataset_id': '5a6446c5-6114-4380-b875-9de17fda2b8d',
                    'dataset_version': '2024_11_07_131440',
                    'dataset_name': 'yt_mixed',
                    'context_speaker_similarity': 0.9172529578208923,
                    'context_audio_offset': 7001.95659375,
                    'context_audio_duration': 14.64,
                    'context_audio_text': 'Uat gives an excellent illustration of the effects of a course of selection, which may be considered as unconscious, insofar that the breeders could never have expected, or even wished, to produce the result which ensued,',
                    'context_recording_id': 'rec-Zdud2gXLTXY',
                },
                alignment=None,
            )
        ],
        features=None,
        recording=Recording(
            id='rec-Zdud2gXLTXY',
            sources=[AudioSource(type='file', channels=[0], source='/audio/Zdud2gXLTXY.wav')],
            sampling_rate=22050,
            num_samples=952064173,
            duration=43177.51351473923,
            channel_ids=[0],
            transforms=None,
        ),
        custom={
            'validation_status': 'pass',
            'target_audio': Recording(
                id='cut-rec-Zdud2gXLTXY-238.16-6.88',
                sources=[AudioSource(type='memory', channels=[0], source='<binary-data>')],
                sampling_rate=22050,
                num_samples=151704,
                duration=6.88,
                channel_ids=[0],
                transforms=None,
            ),
            'context_audio': Recording(
                id='context_cut-rec-Zdud2gXLTXY-7001.96-14.64',
                sources=[AudioSource(type='memory', channels=[0], source='<binary-data>')],
                sampling_rate=22050,
                num_samples=322812,
                duration=14.64,
                channel_ids=[0],
                transforms=None,
            ),
            'target_codes': TemporalArray(
                array=Array(storage_type='memory_npy', storage_path='', storage_key='<binary-data>', shape=[8, 149]),
                temporal_dim=-1,
                frame_shift=0.046511627906976744,
                start=0,
            ),
            'context_codes': TemporalArray(
                array=Array(storage_type='memory_npy', storage_path='', storage_key='<binary-data>', shape=[8, 316]),
                temporal_dim=-1,
                frame_shift=0.046511627906976744,
                start=0,
            ),
            'shard_origin': '/cuts/cuts.000001.jsonl.gz',
            'shar_epoch': 0,
            'tokenizer_names': ['english_phoneme'],
        },
    )
    return cut


def test_cut_cer_filter(cut_example):
    f = CERFilter(0.4)
    assert f(cut_example) == True

    f = CERFilter(0.01)
    assert f(cut_example) == False

    f = CERFilter(float("inf"))
    assert f(cut_example) == True


def test_cut_context_speaker_similarity_filter(cut_example):
    f = ContextSpeakerSimilarityFilter(0.6)
    assert f(cut_example) == True

    f = ContextSpeakerSimilarityFilter(0.95)
    assert f(cut_example) == False

    f = ContextSpeakerSimilarityFilter(-1)
    assert f(cut_example) == True


def test_cut_validation_status_filter(cut_example):
    f = ValidationStatusFilter("pass")
    assert f(cut_example) == True

    f = ValidationStatusFilter("wrong_text")
    assert f(cut_example) == False

    f = ValidationStatusFilter("any_other_status")
    assert f(cut_example) == False



# Copyright (c) 2021, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import contextlib
from typing import List, Optional

import numpy as np
import omegaconf
import torch
import transformers
import wandb
from hydra.utils import instantiate
from lightning.pytorch import Trainer
from lightning.pytorch.loggers import WandbLogger
from omegaconf import DictConfig
from torch import nn
from torch.nn import functional as F
from transformers import AlbertTokenizer

from nemo.collections.common.tokenizers.text_to_speech.tts_tokenizers import (
    EnglishCharsTokenizer,
    EnglishPhonemesTokenizer,
)
from nemo.collections.tts.losses.aligner_loss import BinLoss, ForwardSumLoss
from nemo.collections.tts.models.base import SpectrogramGenerator
from nemo.collections.tts.modules.fastpitch import average_features, regulate_len
from nemo.collections.tts.parts.utils.helpers import (
    binarize_attention_parallel,
    g2p_backward_compatible_support,
    get_mask_from_lengths,
    plot_pitch_to_numpy,
    plot_spectrogram_to_numpy,
)
from nemo.core import Exportable
from nemo.core.classes.common import PretrainedModelInfo, typecheck
from nemo.core.neural_types.elements import (
    LengthsType,
    LogprobsType,
    MelSpectrogramType,
    ProbsType,
    RegressionValuesType,
    TokenDurationType,
    TokenIndex,
    TokenLogDurationType,
)
from nemo.core.neural_types.neural_type import NeuralType
from nemo.utils import logging, model_utils


class MixerTTSModel(SpectrogramGenerator, Exportable):
    """Mixer-TTS and Mixer-TTS-X models (https://arxiv.org/abs/2110.03584) that is used to generate mel spectrogram from text."""

    def __init__(self, cfg: DictConfig, trainer: 'Trainer' = None):
        # Convert to Hydra 1.0 compatible DictConfig
        cfg = model_utils.convert_model_config_to_dict_config(cfg)
        cfg = model_utils.maybe_update_config_version(cfg)

        # Setup normalizer
        self.normalizer = None
        self.text_normalizer_call = None
        self.text_normalizer_call_kwargs = {}
        self._setup_normalizer(cfg)

        # Setup tokenizer
        self.tokenizer = None
        self._setup_tokenizer(cfg)
        assert self.tokenizer is not None

        num_tokens = len(self.tokenizer.tokens)
        self.tokenizer_pad = self.tokenizer.pad
        self.tokenizer_unk = self.tokenizer.oov

        super().__init__(cfg=cfg, trainer=trainer)

        self.pitch_loss_scale = cfg.pitch_loss_scale
        self.durs_loss_scale = cfg.durs_loss_scale
        self.mel_loss_scale = cfg.mel_loss_scale

        self.aligner = instantiate(cfg.alignment_module)
        self.forward_sum_loss = ForwardSumLoss()
        self.bin_loss = BinLoss()
        self.add_bin_loss = False
        self.bin_loss_scale = 0.0
        self.bin_loss_start_ratio = cfg.bin_loss_start_ratio
        self.bin_loss_warmup_epochs = cfg.bin_loss_warmup_epochs

        self.cond_on_lm_embeddings = cfg.get("cond_on_lm_embeddings", False)

        if self.cond_on_lm_embeddings:
            self.lm_padding_value = (
                self._train_dl.dataset.lm_padding_value
                if self._train_dl is not None
                else self._get_lm_padding_value(cfg.lm_model)
            )
            self.lm_embeddings = self._get_lm_embeddings(cfg.lm_model)
            self.lm_embeddings.weight.requires_grad = False

            self.self_attention_module = instantiate(
                cfg.self_attention_module, n_lm_tokens_channels=self.lm_embeddings.weight.shape[1]
            )

        self.encoder = instantiate(cfg.encoder, num_tokens=num_tokens, padding_idx=self.tokenizer_pad)
        self.symbol_emb = self.encoder.to_embed

        self.duration_predictor = instantiate(cfg.duration_predictor)

        self.pitch_mean, self.pitch_std = float(cfg.pitch_mean), float(cfg.pitch_std)
        self.pitch_predictor = instantiate(cfg.pitch_predictor)
        self.pitch_emb = instantiate(cfg.pitch_emb)

        self.preprocessor = instantiate(cfg.preprocessor)

        self.decoder = instantiate(cfg.decoder)
        self.proj = nn.Linear(self.decoder.d_model, cfg.n_mel_channels)

    def _setup_tokenizer(self, cfg):
        text_tokenizer_kwargs = {}
        if "g2p" in cfg.text_tokenizer:
            # for backward compatibility
            if (
                self._is_model_being_restored()
                and (cfg.text_tokenizer.g2p.get('_target_', None) is not None)
                and cfg.text_tokenizer.g2p["_target_"].startswith("nemo_text_processing.g2p")
            ):
                cfg.text_tokenizer.g2p["_target_"] = g2p_backward_compatible_support(
                    cfg.text_tokenizer.g2p["_target_"]
                )

            g2p_kwargs = {}

            if "phoneme_dict" in cfg.text_tokenizer.g2p:
                g2p_kwargs["phoneme_dict"] = self.register_artifact(
                    'text_tokenizer.g2p.phoneme_dict',
                    cfg.text_tokenizer.g2p.phoneme_dict,
                )

            if "heteronyms" in cfg.text_tokenizer.g2p:
                g2p_kwargs["heteronyms"] = self.register_artifact(
                    'text_tokenizer.g2p.heteronyms',
                    cfg.text_tokenizer.g2p.heteronyms,
                )

            text_tokenizer_kwargs["g2p"] = instantiate(cfg.text_tokenizer.g2p, **g2p_kwargs)

        self.tokenizer = instantiate(cfg.text_tokenizer, **text_tokenizer_kwargs)

    def _get_lm_model_tokenizer(self, lm_model="albert"):
        if getattr(self, "_lm_model_tokenizer", None) is not None:
            return self._lm_model_tokenizer

        if self._train_dl is not None and self._train_dl.dataset is not None:
            self._lm_model_tokenizer = self._train_dl.dataset.lm_model_tokenizer

        if lm_model == "albert":
            self._lm_model_tokenizer = AlbertTokenizer.from_pretrained('albert-base-v2')
        else:
            raise NotImplementedError(
                f"{lm_model} lm model is not supported. Only albert is supported at this moment."
            )

        return self._lm_model_tokenizer

    def _get_lm_embeddings(self, lm_model="albert"):
        if lm_model == "albert":
            return transformers.AlbertModel.from_pretrained('albert-base-v2').embeddings.word_embeddings
        else:
            raise NotImplementedError(
                f"{lm_model} lm model is not supported. Only albert is supported at this moment."
            )

    def _get_lm_padding_value(self, lm_model="albert"):
        if lm_model == "albert":
            return transformers.AlbertTokenizer.from_pretrained('albert-base-v2')._convert_token_to_id('<pad>')
        else:
            raise NotImplementedError(
                f"{lm_model} lm model is not supported. Only albert is supported at this moment."
            )

    def _metrics(
        self,
        true_durs,
        true_text_len,
        pred_durs,
        true_pitch,
        pred_pitch,
        true_spect=None,
        pred_spect=None,
        true_spect_len=None,
        attn_logprob=None,
        attn_soft=None,
        attn_hard=None,
        attn_hard_dur=None,
    ):
        text_mask = get_mask_from_lengths(true_text_len)
        mel_mask = get_mask_from_lengths(true_spect_len)
        loss = 0.0

        # Dur loss and metrics
        durs_loss = F.mse_loss(pred_durs, (true_durs + 1).float().log(), reduction='none')
        durs_loss = durs_loss * text_mask.float()
        durs_loss = durs_loss.sum() / text_mask.sum()

        durs_pred = pred_durs.exp() - 1
        durs_pred = torch.clamp_min(durs_pred, min=0)
        durs_pred = durs_pred.round().long()

        acc = ((true_durs == durs_pred) * text_mask).sum().float() / text_mask.sum() * 100
        acc_dist_1 = (((true_durs - durs_pred).abs() <= 1) * text_mask).sum().float() / text_mask.sum() * 100
        acc_dist_3 = (((true_durs - durs_pred).abs() <= 3) * text_mask).sum().float() / text_mask.sum() * 100

        pred_spect = pred_spect.transpose(1, 2)

        # Mel loss
        mel_loss = F.mse_loss(pred_spect, true_spect, reduction='none').mean(dim=-2)
        mel_loss = mel_loss * mel_mask.float()
        mel_loss = mel_loss.sum() / mel_mask.sum()

        loss = loss + self.durs_loss_scale * durs_loss + self.mel_loss_scale * mel_loss

        # Aligner loss
        bin_loss, ctc_loss = None, None
        ctc_loss = self.forward_sum_loss(attn_logprob=attn_logprob, in_lens=true_text_len, out_lens=true_spect_len)
        loss = loss + ctc_loss
        if self.add_bin_loss:
            bin_loss = self.bin_loss(hard_attention=attn_hard, soft_attention=attn_soft)
            loss = loss + self.bin_loss_scale * bin_loss
        true_avg_pitch = average_features(true_pitch.unsqueeze(1), attn_hard_dur).squeeze(1)

        # Pitch loss
        pitch_loss = F.mse_loss(pred_pitch, true_avg_pitch, reduction='none')  # noqa
        pitch_loss = (pitch_loss * text_mask).sum() / text_mask.sum()

        loss = loss + self.pitch_loss_scale * pitch_loss

        return loss, durs_loss, acc, acc_dist_1, acc_dist_3, pitch_loss, mel_loss, ctc_loss, bin_loss

    @torch.jit.unused
    def run_aligner(self, text, text_len, text_mask, spect, spect_len, attn_prior):
        text_emb = self.symbol_emb(text)
        attn_soft, attn_logprob = self.aligner(
            spect,
            text_emb.permute(0, 2, 1),
            mask=text_mask == 0,
            attn_prior=attn_prior,
        )
        attn_hard = binarize_attention_parallel(attn_soft, text_len, spect_len)
        attn_hard_dur = attn_hard.sum(2)[:, 0, :]
        assert torch.all(torch.eq(attn_hard_dur.sum(dim=1), spect_len))
        return attn_soft, attn_logprob, attn_hard, attn_hard_dur

    @typecheck(
        input_types={
            "text": NeuralType(('B', 'T_text'), TokenIndex()),
            "text_len": NeuralType(('B',), LengthsType()),
            "pitch": NeuralType(('B', 'T_audio'), RegressionValuesType(), optional=True),
            "spect": NeuralType(('B', 'D', 'T_spec'), MelSpectrogramType(), optional=True),
            "spect_len": NeuralType(('B',), LengthsType(), optional=True),
            "attn_prior": NeuralType(('B', 'T_spec', 'T_text'), ProbsType(), optional=True),
            "lm_tokens": NeuralType(('B', 'T_lm_tokens'), TokenIndex(), optional=True),
        },
        output_types={
            "pred_spect": NeuralType(('B', 'D', 'T_spec'), MelSpectrogramType()),
            "durs_predicted": NeuralType(('B', 'T_text'), TokenDurationType()),
            "log_durs_predicted": NeuralType(('B', 'T_text'), TokenLogDurationType()),
            "pitch_predicted": NeuralType(('B', 'T_text'), RegressionValuesType()),
            "attn_soft": NeuralType(('B', 'S', 'T_spec', 'T_text'), ProbsType()),
            "attn_logprob": NeuralType(('B', 'S', 'T_spec', 'T_text'), LogprobsType()),
            "attn_hard": NeuralType(('B', 'S', 'T_spec', 'T_text'), ProbsType()),
            "attn_hard_dur": NeuralType(('B', 'T_text'), TokenDurationType()),
        },
    )
    def forward(self, text, text_len, pitch=None, spect=None, spect_len=None, attn_prior=None, lm_tokens=None):
        if self.training:
            assert pitch is not None

        text_mask = get_mask_from_lengths(text_len).unsqueeze(2)

        enc_out, enc_mask = self.encoder(text, text_mask)

        # Aligner
        attn_soft, attn_logprob, attn_hard, attn_hard_dur = None, None, None, None
        if spect is not None:
            attn_soft, attn_logprob, attn_hard, attn_hard_dur = self.run_aligner(
                text, text_len, text_mask, spect, spect_len, attn_prior
            )

        if self.cond_on_lm_embeddings:
            lm_emb = self.lm_embeddings(lm_tokens)
            lm_features = self.self_attention_module(
                enc_out, lm_emb, lm_emb, q_mask=enc_mask.squeeze(2), kv_mask=lm_tokens != self.lm_padding_value
            )

        # Duration predictor
        log_durs_predicted = self.duration_predictor(enc_out, enc_mask)
        durs_predicted = torch.clamp(log_durs_predicted.exp() - 1, 0)

        # Pitch predictor
        pitch_predicted = self.pitch_predictor(enc_out, enc_mask)

        # Avg pitch, add pitch_emb
        if not self.training:
            if pitch is not None:
                pitch = average_features(pitch.unsqueeze(1), attn_hard_dur).squeeze(1)
                pitch_emb = self.pitch_emb(pitch.unsqueeze(1))
            else:
                pitch_emb = self.pitch_emb(pitch_predicted.unsqueeze(1))
        else:
            pitch = average_features(pitch.unsqueeze(1), attn_hard_dur).squeeze(1)
            pitch_emb = self.pitch_emb(pitch.unsqueeze(1))

        enc_out = enc_out + pitch_emb.transpose(1, 2)

        if self.cond_on_lm_embeddings:
            enc_out = enc_out + lm_features

        # Regulate length
        len_regulated_enc_out, dec_lens = regulate_len(attn_hard_dur, enc_out)

        dec_out, dec_lens = self.decoder(len_regulated_enc_out, get_mask_from_lengths(dec_lens).unsqueeze(2))
        pred_spect = self.proj(dec_out)

        return (
            pred_spect,
            durs_predicted,
            log_durs_predicted,
            pitch_predicted,
            attn_soft,
            attn_logprob,
            attn_hard,
            attn_hard_dur,
        )

    def infer(
        self,
        text,
        text_len=None,
        text_mask=None,
        spect=None,
        spect_len=None,
        attn_prior=None,
        use_gt_durs=False,
        lm_tokens=None,
        pitch=None,
    ):
        if text_mask is None:
            text_mask = get_mask_from_lengths(text_len).unsqueeze(2)

        enc_out, enc_mask = self.encoder(text, text_mask)

        # Aligner
        attn_hard_dur = None
        if use_gt_durs:
            attn_soft, attn_logprob, attn_hard, attn_hard_dur = self.run_aligner(
                text, text_len, text_mask, spect, spect_len, attn_prior
            )

        if self.cond_on_lm_embeddings:
            lm_emb = self.lm_embeddings(lm_tokens)
            lm_features = self.self_attention_module(
                enc_out, lm_emb, lm_emb, q_mask=enc_mask.squeeze(2), kv_mask=lm_tokens != self.lm_padding_value
            )

        # Duration predictor
        log_durs_predicted = self.duration_predictor(enc_out, enc_mask)
        durs_predicted = torch.clamp(log_durs_predicted.exp() - 1, 0)

        # Avg pitch, pitch predictor
        if use_gt_durs and pitch is not None:
            pitch = average_features(pitch.unsqueeze(1), attn_hard_dur).squeeze(1)
            pitch_emb = self.pitch_emb(pitch.unsqueeze(1))
        else:
            pitch_predicted = self.pitch_predictor(enc_out, enc_mask)
            pitch_emb = self.pitch_emb(pitch_predicted.unsqueeze(1))

        # Add pitch emb
        enc_out = enc_out + pitch_emb.transpose(1, 2)

        if self.cond_on_lm_embeddings:
            enc_out = enc_out + lm_features

        if use_gt_durs:
            if attn_hard_dur is not None:
                len_regulated_enc_out, dec_lens = regulate_len(attn_hard_dur, enc_out)
            else:
                raise NotImplementedError
        else:
            len_regulated_enc_out, dec_lens = regulate_len(durs_predicted, enc_out)

        dec_out, _ = self.decoder(len_regulated_enc_out, get_mask_from_lengths(dec_lens).unsqueeze(2))
        pred_spect = self.proj(dec_out)

        return pred_spect

    def on_train_epoch_start(self):
        bin_loss_start_epoch = np.ceil(self.bin_loss_start_ratio * self._trainer.max_epochs)

        # Add bin loss when current_epoch >= bin_start_epoch
        if not self.add_bin_loss and self.current_epoch >= bin_loss_start_epoch:
            logging.info(f"Using hard attentions after epoch: {self.current_epoch}")
            self.add_bin_loss = True

        if self.add_bin_loss:
            self.bin_loss_scale = min((self.current_epoch - bin_loss_start_epoch) / self.bin_loss_warmup_epochs, 1.0)

    def training_step(self, batch, batch_idx):
        attn_prior, lm_tokens = None, None
        if self.cond_on_lm_embeddings:
            audio, audio_len, text, text_len, attn_prior, pitch, _, lm_tokens = batch
        else:
            audio, audio_len, text, text_len, attn_prior, pitch, _ = batch

        spect, spect_len = self.preprocessor(input_signal=audio, length=audio_len)

        # pitch normalization
        zero_pitch_idx = pitch == 0
        pitch = (pitch - self.pitch_mean) / self.pitch_std
        pitch[zero_pitch_idx] = 0.0

        (
            pred_spect,
            _,
            pred_log_durs,
            pred_pitch,
            attn_soft,
            attn_logprob,
            attn_hard,
            attn_hard_dur,
        ) = self(
            text=text,
            text_len=text_len,
            pitch=pitch,
            spect=spect,
            spect_len=spect_len,
            attn_prior=attn_prior,
            lm_tokens=lm_tokens,
        )

        (
            loss,
            durs_loss,
            acc,
            acc_dist_1,
            acc_dist_3,
            pitch_loss,
            mel_loss,
            ctc_loss,
            bin_loss,
        ) = self._metrics(
            pred_durs=pred_log_durs,
            pred_pitch=pred_pitch,
            true_durs=attn_hard_dur,
            true_text_len=text_len,
            true_pitch=pitch,
            true_spect=spect,
            pred_spect=pred_spect,
            true_spect_len=spect_len,
            attn_logprob=attn_logprob,
            attn_soft=attn_soft,
            attn_hard=attn_hard,
            attn_hard_dur=attn_hard_dur,
        )

        train_log = {
            'train_loss': loss,
            'train_durs_loss': durs_loss,
            'train_pitch_loss': torch.tensor(1.0).to(durs_loss.device) if pitch_loss is None else pitch_loss,
            'train_mel_loss': mel_loss,
            'train_durs_acc': acc,
            'train_durs_acc_dist_3': acc_dist_3,
            'train_ctc_loss': torch.tensor(1.0).to(durs_loss.device) if ctc_loss is None else ctc_loss,
            'train_bin_loss': torch.tensor(1.0).to(durs_loss.device) if bin_loss is None else bin_loss,
        }

        return {'loss': loss, 'progress_bar': {k: v.detach() for k, v in train_log.items()}, 'log': train_log}

    def validation_step(self, batch, batch_idx):
        attn_prior, lm_tokens = None, None
        if self.cond_on_lm_embeddings:
            audio, audio_len, text, text_len, attn_prior, pitch, _, lm_tokens = batch
        else:
            audio, audio_len, text, text_len, attn_prior, pitch, _ = batch

        spect, spect_len = self.preprocessor(input_signal=audio, length=audio_len)

        # pitch normalization
        zero_pitch_idx = pitch == 0
        pitch = (pitch - self.pitch_mean) / self.pitch_std
        pitch[zero_pitch_idx] = 0.0

        (
            pred_spect,
            _,
            pred_log_durs,
            pred_pitch,
            attn_soft,
            attn_logprob,
            attn_hard,
            attn_hard_dur,
        ) = self(
            text=text,
            text_len=text_len,
            pitch=pitch,
            spect=spect,
            spect_len=spect_len,
            attn_prior=attn_prior,
            lm_tokens=lm_tokens,
        )

        (
            loss,
            durs_loss,
            acc,
            acc_dist_1,
            acc_dist_3,
            pitch_loss,
            mel_loss,
            ctc_loss,
            bin_loss,
        ) = self._metrics(
            pred_durs=pred_log_durs,
            pred_pitch=pred_pitch,
            true_durs=attn_hard_dur,
            true_text_len=text_len,
            true_pitch=pitch,
            true_spect=spect,
            pred_spect=pred_spect,
            true_spect_len=spect_len,
            attn_logprob=attn_logprob,
            attn_soft=attn_soft,
            attn_hard=attn_hard,
            attn_hard_dur=attn_hard_dur,
        )

        # without ground truth internal features except for durations
        pred_spect, _, pred_log_durs, pred_pitch, attn_soft, attn_logprob, attn_hard, attn_hard_dur = self(
            text=text,
            text_len=text_len,
            pitch=None,
            spect=spect,
            spect_len=spect_len,
            attn_prior=attn_prior,
            lm_tokens=lm_tokens,
        )

        *_, with_pred_features_mel_loss, _, _ = self._metrics(
            pred_durs=pred_log_durs,
            pred_pitch=pred_pitch,
            true_durs=attn_hard_dur,
            true_text_len=text_len,
            true_pitch=pitch,
            true_spect=spect,
            pred_spect=pred_spect,
            true_spect_len=spect_len,
            attn_logprob=attn_logprob,
            attn_soft=attn_soft,
            attn_hard=attn_hard,
            attn_hard_dur=attn_hard_dur,
        )

        val_log = {
            'val_loss': loss,
            'val_durs_loss': durs_loss,
            'val_pitch_loss': torch.tensor(1.0).to(durs_loss.device) if pitch_loss is None else pitch_loss,
            'val_mel_loss': mel_loss,
            'val_with_pred_features_mel_loss': with_pred_features_mel_loss,
            'val_durs_acc': acc,
            'val_durs_acc_dist_3': acc_dist_3,
            'val_ctc_loss': torch.tensor(1.0).to(durs_loss.device) if ctc_loss is None else ctc_loss,
            'val_bin_loss': torch.tensor(1.0).to(durs_loss.device) if bin_loss is None else bin_loss,
        }
        self.log_dict(val_log, prog_bar=False, on_epoch=True, logger=True, sync_dist=True)

        if batch_idx == 0 and self.current_epoch % 5 == 0 and isinstance(self.logger, WandbLogger):
            specs = []
            pitches = []
            for i in range(min(3, spect.shape[0])):
                specs += [
                    wandb.Image(
                        plot_spectrogram_to_numpy(spect[i, :, : spect_len[i]].data.cpu().numpy()),
                        caption=f"gt mel {i}",
                    ),
                    wandb.Image(
                        plot_spectrogram_to_numpy(pred_spect.transpose(1, 2)[i, :, : spect_len[i]].data.cpu().numpy()),
                        caption=f"pred mel {i}",
                    ),
                ]

                pitches += [
                    wandb.Image(
                        plot_pitch_to_numpy(
                            average_features(pitch.unsqueeze(1), attn_hard_dur)
                            .squeeze(1)[i, : text_len[i]]
                            .data.cpu()
                            .numpy(),
                            ylim_range=[-2.5, 2.5],
                        ),
                        caption=f"gt pitch {i}",
                    ),
                ]

                pitches += [
                    wandb.Image(
                        plot_pitch_to_numpy(pred_pitch[i, : text_len[i]].data.cpu().numpy(), ylim_range=[-2.5, 2.5]),
                        caption=f"pred pitch {i}",
                    ),
                ]

            self.logger.experiment.log({"specs": specs, "pitches": pitches})

    @typecheck(
        input_types={
            "tokens": NeuralType(('B', 'T_text'), TokenIndex(), optional=True),
            "tokens_len": NeuralType(('B'), LengthsType(), optional=True),
            "lm_tokens": NeuralType(('B', 'T_lm_tokens'), TokenIndex(), optional=True),
            "raw_texts": [NeuralType(optional=True)],
            "lm_model": NeuralType(optional=True),
        },
        output_types={
            "spect": NeuralType(('B', 'D', 'T_spec'), MelSpectrogramType()),
        },
    )
    def generate_spectrogram(
        self,
        tokens: Optional[torch.Tensor] = None,
        tokens_len: Optional[torch.Tensor] = None,
        lm_tokens: Optional[torch.Tensor] = None,
        raw_texts: Optional[List[str]] = None,
        norm_text_for_lm_model: bool = True,
        lm_model: str = "albert",
    ):
        if tokens is not None:
            if tokens_len is None:
                # It is assumed that padding is consecutive and only at the end
                tokens_len = (tokens != self.tokenizer.pad).sum(dim=-1)
        else:
            if raw_texts is None:
                raise ValueError("raw_texts must be specified if tokens is None")

            t_seqs = [self.tokenizer(t) for t in raw_texts]
            tokens = torch.nn.utils.rnn.pad_sequence(
                sequences=[torch.tensor(t, dtype=torch.long, device=self.device) for t in t_seqs],
                batch_first=True,
                padding_value=self.tokenizer.pad,
            )
            tokens_len = torch.tensor([len(t) for t in t_seqs], dtype=torch.long, device=tokens.device)

        if self.cond_on_lm_embeddings and lm_tokens is None:
            if raw_texts is None:
                raise ValueError("raw_texts must be specified if lm_tokens is None")

            lm_model_tokenizer = self._get_lm_model_tokenizer(lm_model)
            lm_padding_value = lm_model_tokenizer._convert_token_to_id('<pad>')
            lm_space_value = lm_model_tokenizer._convert_token_to_id('▁')

            assert isinstance(self.tokenizer, EnglishCharsTokenizer) or isinstance(
                self.tokenizer, EnglishPhonemesTokenizer
            )

            if norm_text_for_lm_model and self.text_normalizer_call is not None:
                raw_texts = [self.text_normalizer_call(t, **self.text_normalizer_call_kwargs) for t in raw_texts]

            preprocess_texts_as_tts_input = [self.tokenizer.text_preprocessing_func(t) for t in raw_texts]
            lm_tokens_as_ids_list = [
                lm_model_tokenizer.encode(t, add_special_tokens=False) for t in preprocess_texts_as_tts_input
            ]

            if self.tokenizer.pad_with_space:
                lm_tokens_as_ids_list = [[lm_space_value] + t + [lm_space_value] for t in lm_tokens_as_ids_list]

            lm_tokens = torch.full(
                (len(lm_tokens_as_ids_list), max([len(t) for t in lm_tokens_as_ids_list])),
                fill_value=lm_padding_value,
                device=tokens.device,
            )
            for i, lm_tokens_i in enumerate(lm_tokens_as_ids_list):
                lm_tokens[i, : len(lm_tokens_i)] = torch.tensor(lm_tokens_i, device=tokens.device)

        pred_spect = self.infer(tokens, tokens_len, lm_tokens=lm_tokens).transpose(1, 2)
        return pred_spect

    def parse(self, text: str, normalize=True) -> torch.Tensor:
        if self.training:
            logging.warning("parse() is meant to be called in eval mode.")
        if normalize and self.text_normalizer_call is not None:
            text = self.text_normalizer_call(text, **self.text_normalizer_call_kwargs)

        eval_phon_mode = contextlib.nullcontext()
        if hasattr(self.tokenizer, "set_phone_prob"):
            eval_phon_mode = self.tokenizer.set_phone_prob(prob=1.0)

        with eval_phon_mode:
            tokens = self.tokenizer.encode(text)
        return torch.tensor(tokens).long().unsqueeze(0).to(self.device)

    def _loader(self, cfg):
        try:
            _ = cfg.dataset.manifest_filepath
        except omegaconf.errors.MissingMandatoryValue:
            logging.warning("manifest_filepath was skipped. No dataset for this model.")
            return None

        dataset = instantiate(
            cfg.dataset,
            text_normalizer=self.normalizer,
            text_normalizer_call_kwargs=self.text_normalizer_call_kwargs,
            text_tokenizer=self.tokenizer,
        )
        return torch.utils.data.DataLoader(  # noqa
            dataset=dataset,
            collate_fn=dataset.collate_fn,
            **cfg.dataloader_params,
        )

    def setup_training_data(self, cfg):
        self._train_dl = self._loader(cfg)

    def setup_validation_data(self, cfg):
        self._validation_dl = self._loader(cfg)

    def setup_test_data(self, cfg):
        """Omitted."""
        pass

    @classmethod
    def list_available_models(cls) -> 'List[PretrainedModelInfo]':
        """
        This method returns a list of pre-trained model which can be instantiated directly from NVIDIA's NGC cloud.
        Returns:
            List of available pre-trained models.
        """
        list_of_models = []
        model = PretrainedModelInfo(
            pretrained_model_name="tts_en_lj_mixertts",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/tts_en_lj_mixertts/versions/1.6.0/files/tts_en_lj_mixertts.nemo",
            description="This model is trained on LJSpeech sampled at 22050Hz with and can be used to generate female English voices with an American accent.",
            class_=cls,  # noqa
        )
        list_of_models.append(model)

        model = PretrainedModelInfo(
            pretrained_model_name="tts_en_lj_mixerttsx",
            location="https://api.ngc.nvidia.com/v2/models/nvidia/nemo/tts_en_lj_mixerttsx/versions/1.6.0/files/tts_en_lj_mixerttsx.nemo",
            description="This model is trained on LJSpeech sampled at 22050Hz with and can be used to generate female English voices with an American accent.",
            class_=cls,  # noqa
        )
        list_of_models.append(model)

        return list_of_models

    # Methods for model exportability
    @property
    def input_types(self):
        return {
            "text": NeuralType(('B', 'T_text'), TokenIndex()),
            "lm_tokens": NeuralType(('B', 'T_lm_tokens'), TokenIndex(), optional=True),
        }

    @property
    def output_types(self):
        return {
            "spect": NeuralType(('B', 'D', 'T_spec'), MelSpectrogramType()),
        }

    def input_example(self, max_text_len=10, max_lm_tokens_len=10):
        text = torch.randint(
            low=0,
            high=len(self.tokenizer.tokens),
            size=(1, max_text_len),
            device=self.device,
            dtype=torch.long,
        )

        inputs = {'text': text}

        if self.cond_on_lm_embeddings:
            inputs['lm_tokens'] = torch.randint(
                low=0,
                high=self.lm_embeddings.weight.shape[0],
                size=(1, max_lm_tokens_len),
                device=self.device,
                dtype=torch.long,
            )

        return (inputs,)

    def forward_for_export(self, text, lm_tokens=None):
        text_mask = (text != self.tokenizer_pad).unsqueeze(2)
        spect = self.infer(text=text, text_mask=text_mask, lm_tokens=lm_tokens).transpose(1, 2)
        return spect.to(torch.float)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import torch

from nemo.collections.speechlm2.models.salm import replace_placeholders_and_build_targets


def test_replace_placeholders():
    # fmt: off
    PAD = 0
    AUDIO = 100
    input_ids = torch.tensor([
        [7  , AUDIO, 1, 2    , AUDIO, 1],
        [PAD,   PAD, 3, AUDIO, 4    , 5]  # note: left padding required
    ])
    loss_mask = torch.tensor([
        [False, False, False, False, False, True],  # predict last token
        [False, False, False, False, True , True]  # predict last two tokens
    ])
    embeds = torch.ones(2, 6, 2)
    embeds[1, :2] = 0  # note: indicate left padding
    # 3 embedding sequences with varying shapes, corresponding to 3 AUDIO tokens
    replacements = [
        torch.full((4, 2), fill_value=2.0),
        torch.full((3, 2), fill_value=3.0),
        torch.full((2, 2), fill_value=4.0),
    ]

    embeds_r, targets_r, attention_mask_r = replace_placeholders_and_build_targets(
        input_ids=input_ids,
        embeds=embeds,
        padding_id=PAD,
        placeholder_id=AUDIO,
        replacements=replacements,
        target_ids=input_ids.where(loss_mask, -100)
    )

    assert embeds_r.shape == (2, 11, 2)
    # batch item 0
    assert (embeds_r[0, 0]    == 1.0).all()  # 1=orig
    assert (embeds_r[0, 1:5]  == 2.0).all()  # 2=repl
    assert (embeds_r[0, 5:7]  == 1.0).all()  # 1=orig
    assert (embeds_r[0, 7:10] == 3.0).all()  # 3=repl
    assert (embeds_r[0, 10]   == 1.0).all()  # 1=orig
    # batch item 1
    assert (embeds_r[1, :6]   == 0.0).all()  # 0=pad
    assert (embeds_r[1, 6:7]  == 1.0).all()  # 1=orig
    assert (embeds_r[1, 7:9]  == 4.0).all()  # 4=repl
    assert (embeds_r[1, 9:]   == 1.0).all()  # 1=orig

    assert targets_r.shape == (2, 11)
    torch.testing.assert_close(
        targets_r,
        torch.tensor([
            [-100, -100, -100, -100, -100, -100, -100, -100, -100, -100, 1],
            [-100, -100, -100, -100, -100, -100, -100, -100, -100, 4   , 5],
        ])
    )

    assert attention_mask_r.shape == (2, 11)
    torch.testing.assert_close(
        attention_mask_r,
        torch.tensor([
            [True, True, True, True, True,  True,  True,  True,  True,  True,  True],
            [False, False, False, False, False, False, True, True, True, True, True],
        ])
    )
    # fmt: on


def test_replace_placeholders_removes_excessive_left_padding():
    # fmt: off
    PAD = 0
    AUDIO = 100
    input_ids = torch.tensor([
        [7  , AUDIO, 1    , 2],
        [PAD,   PAD, AUDIO, 3]  # note: left padding required
    ])
    loss_mask = torch.tensor([
        [False, False, True , True],  # predict last two tokens
        [False, False, False, True]   # predict last token
    ])
    embeds = torch.ones(2, 4, 2)
    embeds[1, :2] = 0  # note: indicate left padding
    # 3 embedding sequences with varying shapes, corresponding to 3 AUDIO tokens
    replacements = [
        torch.full((3, 2), fill_value=2.0),
        torch.full((5, 2), fill_value=4.0),
    ]

    embeds_r, targets_r, attention_mask_r = replace_placeholders_and_build_targets(
        input_ids=input_ids,
        embeds=embeds,
        padding_id=PAD,
        placeholder_id=AUDIO,
        replacements=replacements,
        target_ids=input_ids.where(loss_mask, -100)
    )

    assert embeds_r.shape == (2, 6, 2)
    # batch item 0
    assert (embeds_r[0, 0  ]  == 1.0).all()  # 1=orig
    assert (embeds_r[0, 1:4]  == 2.0).all()  # 2=repl
    assert (embeds_r[0, 4: ]  == 1.0).all()  # 1=orig
    # batch item 1
    assert (embeds_r[1, :5]  == 4.0).all()  # 4=repl
    assert (embeds_r[1, 5 ]  == 1.0).all()  # 1=orig

    assert targets_r.shape == (2, 6)
    torch.testing.assert_close(
        targets_r,
        torch.tensor([
            [-100, -100, -100, -100,    1, 2],
            [-100, -100, -100, -100, -100, 3],
        ])
    )

    assert attention_mask_r.shape == (2, 6)
    torch.testing.assert_close(
        attention_mask_r,
        torch.tensor([
            [True, True, True, True, True, True],
            [True, True, True, True, True, True],
        ])
    )
    # fmt: on



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from nemo.lightning.io.artifact import DirOrStringArtifact, FileArtifact
from nemo.lightning.io.mixin import track_io

# Registers all required classes with track_io functionality
try:
    # Track HF tokenizers
    from transformers import AutoTokenizer as HfAutoTokenizer
    from transformers.models.llama.tokenization_llama import LlamaTokenizer
    from transformers.models.llama.tokenization_llama_fast import LlamaTokenizerFast

    for cls in [HfAutoTokenizer, LlamaTokenizer, LlamaTokenizerFast]:
        track_io(
            cls,
            artifacts=[
                FileArtifact(attr_name, required=False)
                for attr_name in ['vocab_file', 'merges_file', 'tokenizer_file', 'name_or_path']
            ],
        )

    from nemo.collections.common.tokenizers import AutoTokenizer

    track_io(
        AutoTokenizer,
        artifacts=[
            FileArtifact("vocab_file", required=False),
            FileArtifact("merges_file", required=False),
            DirOrStringArtifact("pretrained_model_name", required=False),
        ],
    )
except ImportError:
    # HF tokenizers are not available, no need to track them
    pass


try:
    from nemo.collections.common.tokenizers import ByteLevelTokenizer, SentencePieceTokenizer, TiktokenTokenizer

    track_io(SentencePieceTokenizer, artifacts=[FileArtifact("model_path")])
    track_io(TiktokenTokenizer, artifacts=[FileArtifact("vocab_file")])
    track_io(ByteLevelTokenizer)
except ImportError:
    # Tokenizers are not available, no need to track it.
    pass



# Copyright (c) 2022, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import multiprocessing
import os
from dataclasses import dataclass, field
from pathlib import Path

import pytest
from hydra.utils import instantiate
from omegaconf import OmegaConf

try:
    from nemo_text_processing.text_normalization.normalize import Normalizer
except (ImportError, ModuleNotFoundError):
    raise ModuleNotFoundError(
        "The package `nemo_text_processing` was not installed in this environment. Please refer to"
        " https://github.com/NVIDIA/NeMo-text-processing and install this package before using "
        "this script"
    )

from nemo.collections.asr.data.text_to_text import TextToTextDataset, TextToTextItem, TextToTextIterableDataset
from nemo.collections.common import tokenizers

BASE_DIR = Path(__file__).parent.parent.parent.parent


@pytest.fixture(scope="module")
def set_multiprocessing_method():
    """
    Try to set 'fork' multiprocessing method to avoid problems with multiprocessing in PyTest on MacOS
    """
    if multiprocessing.get_start_method(allow_none=True) != "fork":
        multiprocessing.set_start_method("fork", force=True)


@pytest.fixture(scope="module")
def speakers_path(tmp_path_factory):
    path = tmp_path_factory.mktemp("textonly") / "speakers.txt"
    with open(path, "w", encoding="utf-8") as f:
        for speaker in [1, 2, 3]:
            print(f"{speaker}", file=f)
    return path


@pytest.fixture(scope="module")
def textonly_manifest_path(tmp_path_factory):
    path = tmp_path_factory.mktemp("textonly") / "manifest.json"
    texts = [
        "lorem ipsum dolor sit amet consectetur adipiscing elit",
        "nullam rhoncus sapien eros eu mollis sem euismod non",
    ]
    with open(path, "w", encoding="utf-8") as f:
        for text in texts:
            print(json.dumps(dict(text=text, tts_text_normalized=text)), file=f)
    return path


@pytest.fixture(scope="module")
def textonly_unnormalized_manifest_path(tmp_path_factory):
    path = tmp_path_factory.mktemp("textonly") / "manifest_nonorm.json"
    texts = [
        (
            "lorem ipsum dolor sit amet consectetur adipiscing elit",
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit.",
        ),
        (
            "nullam rhoncus sapien eros eu mollis sem euismod non nineteen",
            "Nullam rhoncus sapien eros, eu mollis sem euismod non 19.",
        ),
    ]
    with open(path, "w", encoding="utf-8") as f:
        for asr_text, tts_text in texts:
            print(json.dumps(dict(text=asr_text, tts_text=tts_text)), file=f)
    return path


@pytest.fixture(scope="module")
def tts_normalizer():
    normalizer = Normalizer(lang="en", input_case="cased", overwrite_cache=True, cache_dir=None,)
    return normalizer


@pytest.fixture(scope="module")
def asr_tokenizer(test_data_dir):
    tokenizer_path = os.path.join(test_data_dir, "asr", "tokenizers", "an4_wpe_128", 'vocab.txt')
    tokenizer = tokenizers.AutoTokenizer(pretrained_model_name='bert-base-cased', vocab_file=tokenizer_path)
    return tokenizer


@pytest.fixture(scope="module")
def tts_tokenizer():
    @dataclass
    class G2PConfig:
        _target_: str = "nemo.collections.tts.g2p.models.en_us_arpabet.EnglishG2p"
        phoneme_dict: str = str(BASE_DIR / "scripts/tts_dataset_files/cmudict-0.7b_nv22.10")
        heteronyms: str = str(BASE_DIR / "scripts/tts_dataset_files/heteronyms-052722")
        phoneme_probability: float = 0.5

    @dataclass
    class TextTokenizerCfg:
        _target_: str = "nemo.collections.common.tokenizers.text_to_speech.tts_tokenizers.EnglishPhonemesTokenizer"
        punct: bool = True
        stresses: bool = True
        chars: bool = True
        apostrophe: bool = True
        pad_with_space: bool = True
        add_blank_at: bool = True
        g2p: G2PConfig = field(default_factory=lambda: G2PConfig())

    config = OmegaConf.create(OmegaConf.to_yaml(TextTokenizerCfg()))
    return instantiate(config)


class TestTextToTextDataset:
    @pytest.mark.unit
    @pytest.mark.parametrize("tokenizer_workers", [1, 2])
    def test_text_to_text_dataset(
        self,
        textonly_manifest_path,
        tokenizer_workers,
        speakers_path,
        asr_tokenizer,
        tts_tokenizer,
        tts_normalizer,
        set_multiprocessing_method,
    ):
        """
        Test map-style text-to-text dataset with ASR and TTS tokenizers with normalized text
        """
        dataset = TextToTextDataset(
            manifest_filepath=textonly_manifest_path,
            speakers_filepath=speakers_path,
            asr_tokenizer=asr_tokenizer,
            asr_use_start_end_token=False,
            tts_parser=tts_tokenizer,
            tts_text_pad_id=0,
            tts_text_normalizer=tts_normalizer,
            tts_text_normalizer_call_kwargs=dict(),
            tokenizer_workers=tokenizer_workers,
        )
        assert len(dataset) == 2
        item = dataset[0]
        assert isinstance(item, TextToTextItem)

    @pytest.mark.unit
    def test_text_to_text_dataset_unnormalized(
        self, textonly_unnormalized_manifest_path, speakers_path, asr_tokenizer, tts_tokenizer, tts_normalizer
    ):
        """
        Test TextToTextDataset with ASR and TTS tokenizers with non-normalized text
        """
        dataset = TextToTextDataset(
            manifest_filepath=textonly_unnormalized_manifest_path,
            speakers_filepath=speakers_path,
            asr_tokenizer=asr_tokenizer,
            asr_use_start_end_token=False,
            tts_parser=tts_tokenizer,
            tts_text_pad_id=0,
            tts_text_normalizer=tts_normalizer,
            tts_text_normalizer_call_kwargs=dict(),
        )
        assert len(dataset) == 2

    @pytest.mark.unit
    @pytest.mark.parametrize("tokenizer_workers", [1, 2])
    def test_text_to_text_iterable_dataset(
        self,
        textonly_manifest_path,
        tokenizer_workers,
        speakers_path,
        asr_tokenizer,
        tts_tokenizer,
        tts_normalizer,
        set_multiprocessing_method,
    ):
        """
        Test iterable text-to-text dataset with ASR and TTS tokenizers with normalized text
        """
        dataset = TextToTextIterableDataset(
            manifest_filepath=textonly_manifest_path,
            speakers_filepath=speakers_path,
            asr_tokenizer=asr_tokenizer,
            asr_use_start_end_token=False,
            tts_parser=tts_tokenizer,
            tts_text_pad_id=0,
            tts_text_normalizer=tts_normalizer,
            tts_text_normalizer_call_kwargs=dict(),
            tokenizer_workers=tokenizer_workers,
        )
        assert len(dataset) == 2
        item = next(iter(dataset))
        assert isinstance(item, TextToTextItem)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from omegaconf import DictConfig

from nemo.collections.asr.inference.streaming.framing.request import FeatureBuffer
from nemo.collections.asr.inference.utils.constants import LOG_MEL_ZERO


class FeatureBufferer:
    """
    Feature bufferer class
    It buffers the feature chunks and maintains the buffer.
    """

    def __init__(
        self,
        sample_rate: int,
        buffer_size_in_secs: float,
        preprocessor_cfg: DictConfig,
        device: torch.device,
        fill_value: float = LOG_MEL_ZERO,
    ):
        """
        Args:
            sample_rate (int): sample rate
            buffer_size_in_secs (float): buffer size in seconds
            preprocessor_cfg (DictConfig): preprocessor config
            device (torch.device): device
            fill_value (float): value to fill the feature buffer with
        """
        self.sample_rate = sample_rate
        self.buffer_size_in_secs = buffer_size_in_secs
        self.device = device

        if hasattr(preprocessor_cfg, 'log') and preprocessor_cfg.log:
            self.ZERO_LEVEL_SPEC_DB_VAL = LOG_MEL_ZERO
        else:
            self.ZERO_LEVEL_SPEC_DB_VAL = fill_value

        self.n_feat = preprocessor_cfg.features
        self.feature_buffer_len = int(buffer_size_in_secs / preprocessor_cfg.window_stride)
        self.feature_buffer = torch.full(
            [self.n_feat, self.feature_buffer_len],
            self.ZERO_LEVEL_SPEC_DB_VAL,
            dtype=torch.float32,
            device=self.device,
        )

    def reset(self) -> None:
        """
        Reset the buffer to zero
        """
        self.feature_buffer.fill_(self.ZERO_LEVEL_SPEC_DB_VAL)

    def update(self, fbuffer: FeatureBuffer) -> None:
        """
        Replace feature buffer with new data
        Args:
            fbuffer (FeatureBuffer): feature buffer to update
        """
        # Resize if needed (optional)
        if fbuffer.size != self.feature_buffer.shape[1]:
            self.feature_buffer = torch.full(
                [self.n_feat, fbuffer.size],
                self.ZERO_LEVEL_SPEC_DB_VAL,
                dtype=torch.float32,
                device=self.device,
            )

        self.feature_buffer.copy_(fbuffer.features)

    def get_feature_buffer(self) -> torch.Tensor:
        """
        Get the current feature buffer
        Returns:
            torch.Tensor: current state of the feature buffer
        """
        return self.feature_buffer.clone()


class BatchedFeatureBufferer:
    """
    Batched feature bufferer class
    It buffers the feature chunks from multiple streams and maintains the buffers.
    """

    def __init__(
        self,
        sample_rate: int,
        buffer_size_in_secs: float,
        preprocessor_cfg: DictConfig,
        device: torch.device,
    ):
        """
        Args:
            sample_rate (int): sample rate
            buffer_size_in_secs (float): buffer size in seconds
            preprocessor_cfg (DictConfig): preprocessor config
            device (torch.device): device
        """
        self.sample_rate = sample_rate
        self.buffer_size_in_secs = buffer_size_in_secs
        self.preprocessor_cfg = preprocessor_cfg
        self.device = device
        self.bufferers = {}

    def reset(self) -> None:
        """Reset bufferers"""
        self.bufferers = {}

    def rm_bufferer(self, stream_id: int) -> None:
        """
        Remove bufferer for the given stream id
        Args:
            stream_id (int): stream id
        """
        self.bufferers.pop(stream_id, None)

    def update(self, fbuffers: list[FeatureBuffer]) -> list[torch.Tensor]:
        """
        Update the feature bufferers with the new feature buffers.
        Feature buffers can come from different streams (audios), so we need to maintain a bufferer for each stream.
        Args:
            fbuffers (list[FeatureBuffer]): list of feature buffers
        Returns:
            list[torch.Tensor]: list of feature buffers, one per input frame
        """
        result_buffers = []
        for fbuffer in fbuffers:
            bufferer = self.bufferers.get(fbuffer.stream_id, None)

            if bufferer is None:
                bufferer = FeatureBufferer(
                    self.sample_rate,
                    self.buffer_size_in_secs,
                    self.preprocessor_cfg,
                    self.device,
                )
                self.bufferers[fbuffer.stream_id] = bufferer

            bufferer.update(fbuffer)
            result_buffers.append(bufferer.get_feature_buffer())

            if fbuffer.is_last:
                self.rm_bufferer(fbuffer.stream_id)

        return result_buffers
